"""
The MatchInfoType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from xml.etree import ElementTree
from typing import List, Union, Dict, Optional

from sarpy.io.xml.base import Serializable, ParametersCollection, \
    get_node_value, find_first_child, find_children
from sarpy.io.xml.descriptors import StringDescriptor, IntegerDescriptor, \
    SerializableListDescriptor, ParametersDescriptor

from .base import DEFAULT_STRICT


class MatchCollectionType(Serializable):
    """The match collection type."""
    _fields = ('CoreName', 'MatchIndex', 'Parameters')
    _required = ('CoreName', )
    _collections_tags = {'Parameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    CoreName = StringDescriptor(
        'CoreName', _required, strict=DEFAULT_STRICT,
        docstring='Unique identifier for the match type.')  # type: str
    MatchIndex = IntegerDescriptor(
        'MatchIndex', _required, strict=DEFAULT_STRICT,
        docstring='Collection sequence index for the match collection, assuming '
                  'that this makes sense.')  # type: int
    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The match parameters.')  # type: ParametersCollection

    def __init__(
            self,
            CoreName: str = None,
            MatchIndex: Optional[int] = None,
            Parameters: Union[None, ParametersCollection, Dict] = None,
            **kwargs):
        """

        Parameters
        ----------
        CoreName : str
        MatchIndex : int
        Parameters : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CoreName = CoreName
        self.MatchIndex = MatchIndex
        self.Parameters = Parameters
        super(MatchCollectionType, self).__init__(**kwargs)


class MatchType(Serializable):
    """The is an array element for match information."""
    _fields = ('TypeID', 'CurrentIndex', 'NumMatchCollections', 'MatchCollections')
    _required = ('TypeID', 'NumMatchCollections')
    _collections_tags = {'MatchCollections': {'array': False, 'child_tag': 'MatchCollection'}}
    # descriptors
    TypeID = StringDescriptor(
        'TypeID', _required, strict=DEFAULT_STRICT,
        docstring='The match type identifier. *Examples - "MULTI-IMAGE", "COHERENT" or "STEREO"*')  # type: str
    CurrentIndex = IntegerDescriptor(
        'CurrentIndex', _required, strict=DEFAULT_STRICT,
        docstring='Collection sequence index for the current collection. That is, which collection in the '
                  'collection series (defined in MatchCollections) is this collection? '
                  '(1-based enumeration).')  # type: Optional[int]
    MatchCollections = SerializableListDescriptor(
        'MatchCollections', MatchCollectionType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The match collections.')  # type: Optional[List[MatchCollectionType]]

    def __init__(
            self,
            TypeID: str = None,
            CurrentIndex: Optional[int] = None,
            MatchCollections: Optional[List[MatchCollectionType]] = None,
            **kwargs):
        """

        Parameters
        ----------
        TypeID : str
        CurrentIndex : None|int
        MatchCollections : None|List[MatchCollectionType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TypeID = TypeID
        self.CurrentIndex = CurrentIndex
        self.MatchCollections = MatchCollections
        super(MatchType, self).__init__(**kwargs)

    @property
    def NumMatchCollections(self):
        """int: The number of match collections for this match type."""
        if self.MatchCollections is None:
            return 0
        else:
            return len(self.MatchCollections)


class MatchInfoType(Serializable):
    """
    The match information container. This contains data for multiple collection taskings.
    """

    _fields = ('NumMatchTypes', 'MatchTypes')
    _required = _fields
    _collections_tags = {'MatchTypes': {'array': False, 'child_tag': 'MatchType'}}
    # descriptors
    MatchTypes = SerializableListDescriptor(
        'MatchTypes', MatchType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The match types list.')  # type: List[MatchType]

    def __init__(
            self,
            MatchTypes: List[MatchType] = None,
            **kwargs):
        """

        Parameters
        ----------
        MatchTypes : List[MatchType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.MatchTypes = MatchTypes
        super(MatchInfoType, self).__init__(**kwargs)

    @property
    def NumMatchTypes(self) -> int:
        """int: The number of types of matched collections."""
        if self.MatchTypes is None:
            return 0
        else:
            return len(self.MatchTypes)

    @classmethod
    def _from_node_0_5(cls, node, xml_ns, ns_key):
        """
        Helper method, not really for public usage. For XML deserialization from SICD version prior to 1.0.

        Parameters
        ----------
        node : ElementTree.Element
            dom element for serialized class instance
        xml_ns : dict
            The xml namespace dictionary
        ns_key : str
            The namespace key in the dictionary
        Returns
        -------
        Serializable
            corresponding class instance
        """

        def get_element(tid, cid, cname, params):
            return {
                'TypeID': tid,
                'CurrentIndex': cid,
                'MatchCollections': [{'CoreName': cname, 'Parameters': params}, ]}

        # Note that this is NOT converting the MatchType.MatchCollection in spirit.
        # There isn't enough structure to guarantee that you actually can. This will
        # always yield MatchType.MatchCollection length 1, because the collection details are stuffed
        # into the parameters free form, while CurrentIndex is extracted and actually yields the
        # collection index number (likely larger than 1). This is at least confusing, but more likely
        # completely misleading.
        match_types = []

        coll_key = cls._child_xml_ns_key.get('Collect', ns_key)
        cnodes = find_children(node, 'Collect', xml_ns, coll_key)
        for cnode in cnodes:  # assumed non-empty
            # this describes one series of collects, possibly with more than one MatchType = TypeID
            # It is not clear how it would be possible to deconflict a repeat of MatchType between
            # Collect tags, so I will not.
            core_key = cls._child_xml_ns_key.get('CoreName', ns_key)
            core_name = get_node_value(find_first_child(cnode, 'CoreName', xml_ns, core_key))
            current_index = None
            parameters = []

            pkey = cls._child_xml_ns_key.get('Parameters', ns_key)
            pnodes = find_children(cnode, 'Parameter', xml_ns, pkey)
            for pnode in pnodes:
                name = pnode.attrib['name']
                value = get_node_value(pnode)
                if name == 'CURRENT_INSTANCE':
                    current_index = int(value)  # extract the current index (and exclude)
                else:
                    parameters.append({'name': name, 'value': value})  # copy the parameter
            if current_index is None:
                continue  # I don't know what we would do?
            mt_key = cls._child_xml_ns_key.get('MatchType', ns_key)
            mtypes = find_children(cnode, 'MatchType', xml_ns, mt_key)
            for tnode in mtypes:
                type_id = get_node_value(tnode)
                match_types.append(get_element(type_id, current_index, core_name, parameters))
        if len(match_types) > 0:
            # noinspection PyTypeChecker
            return cls(MatchTypes=match_types)
        else:
            return None

    @classmethod
    def from_node(cls, node, xml_ns, ns_key=None, kwargs=None):
        coll_key = cls._child_xml_ns_key.get('Collect', ns_key)
        coll = find_first_child(node, 'Collect', xml_ns, coll_key)
        if coll is not None:
            # This is from SICD version prior to 1.0, so handle manually.
            return cls._from_node_0_5(node, xml_ns, ns_key)
        else:
            return super(MatchInfoType, cls).from_node(node, xml_ns, ns_key=ns_key, kwargs=kwargs)



"""
Module providing api consistent with other file types for reading tiff files.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Daniel Haverporth")

# It was the original intent to use gdal for the bulk of tiff reading
# Unfortunately, the necessary sarpy functionality can only be obtained by
# gdal_dataset.GetVirtualMemArray(). As of July 2020, this is supported only
# on Linux platforms - unclear what more constraints. So, using gdal to provide
# the reading capability is not feasible at present.

import logging
import os

import numpy
import re
from typing import Union, Tuple, Dict, BinaryIO, Sequence

from sarpy.io.general.base import BaseReader, SarpyIOError
from sarpy.io.general.format_function import ComplexFormatFunction
from sarpy.io.general.data_segment import NumpyMemmapSegment

logger = logging.getLogger(__name__)


_BASELINE_TAGS = {
    254: 'NewSubfileType',
    255: 'SubfileType',
    256: 'ImageWidth',
    257: 'ImageLength',
    258: 'BitsPerSample',
    259: 'Compression',
    262: 'PhotometricInterpretation',
    263: 'Thresholding',
    264: 'CellWidth',
    265: 'CellLength',
    266: 'FillOrder',
    270: 'ImageDescription',
    271: 'Make',
    272: 'Model',
    273: 'StripOffsets',
    274: 'Orientation',
    277: 'SamplesPerPixel',
    278: 'RowsPerStrip',
    279: 'StripByteCounts',
    280: 'MinSampleValue',
    281: 'MaxSampleValue',
    282: 'XResolution',
    283: 'YResolution',
    284: 'PlanarConfiguration',
    288: 'FreeOffsets',
    289: 'FreeByteCounts',
    290: 'GrayResponseUnit',
    291: 'GrayResponseCurve',
    296: 'ResolutionUnit',
    305: 'Software',
    306: 'DateTime',
    315: 'Artist',
    316: 'HostComputer',
    320: 'ColorMap',
    338: 'ExtraSamples',
    33432: 'Copyright',
}
_EXTENSION_TAGS = {
    269: 'DocumentName',
    285: 'PageName',
    286: 'XPosition',
    287: 'YPosition',
    292: 'T4Options',
    293: 'T6Options',
    297: 'PageNumber',
    301: 'TransferFunction',
    317: 'Predictor',
    318: 'WhitePoint',
    319: 'PrimaryChromaticities',
    321: 'HalftoneHints',
    322: 'TileWidth',
    323: 'TileLength',
    324: 'TileOffsets',
    325: 'TileByteCounts',
    326: 'BadFaxLines',
    327: 'CleanFaxData',
    328: 'ConsecutiveBadFaxLines',
    330: 'SubIFDs',
    332: 'InkSet',
    333: 'InkNames',
    334: 'NumberOfInks',
    336: 'DotRange',
    337: 'TargetPrinter',
    339: 'SampleFormat',
    340: 'SMinSampleValue',
    341: 'SMaxSampleValue',
    342: 'TransferRange',
    343: 'ClipPath',
    344: 'XClipPathUnits',
    345: 'YClipPathUnits',
    346: 'Indexed',
    347: 'JPEGTables',
    351: 'OPIProxy',
    400: 'GlobalParametersIFD',
    401: 'ProfileType',
    402: 'FaxProfile',
    403: 'CodingMethods',
    404: 'VersionYear',
    405: 'ModeNumber',
    433: 'Decode',
    434: 'DefaultImageColor',
    512: 'JPEGProc',
    513: 'JPEGInterchangeFormat',
    514: 'JPEGInterchangeFormatLength',
    515: 'JPEGRestartInterval',
    517: 'JPEGLosslessPredictors',
    518: 'JPEGPointTransforms',
    519: 'JPEGQTables',
    520: 'JPEGDCTables',
    521: 'JPEGACTables',
    529: 'YCbCrCoefficients',
    530: 'YCbCrSubSampling',
    531: 'YCbCrPositioning',
    532: 'ReferenceBlackWhite',
    559: 'StripRowCounts',
    700: 'XMP',
    32781: 'ImageID',
    34732: 'ImageLayer',
}
_GEOTIFF_TAGS = {
    33550: 'ModelPixelScaleTag',
    33922: 'ModelTiePointTag',
    34264: 'ModelTransformationTag',
    34735: 'GeoKeyDirectoryTag',
    34736: 'GeoDoubleParamsTag',
    34737: 'GeoAsciiParamsTag',
}


##########

class TiffDetails(object):
    """
    For checking tiff metadata, and parsing in the event we are not using GDAL
    """

    __slots__ = ('_file_name', '_endian', '_magic_number', '_tags')
    _DTYPES = {i+1: entry for i, entry in enumerate(
        ['u1', 'a', 'u2', 'u4', 'u4',
         'i1', 'u1', 'i2', 'i4', 'i4',
         'f4', 'f8', 'u4', None, None,
         'u8', 'i8', 'u8'])}
    _SIZES = numpy.array(
        [1, 1, 2, 4, 8,
         1, 1, 2, 4, 8,
         4, 8, 4, 0, 0,
         8, 8, 8], dtype=numpy.int64)
    # no definition for entries for 14 & 15

    def __init__(self, file_name: str):
        """

        Parameters
        ----------
        file_name : str
        """

        if not (isinstance(file_name, str) and os.path.isfile(file_name)):
            raise SarpyIOError('Not a TIFF file.')

        with open(file_name, 'rb') as fi:
            # Try to read the basic tiff header
            try:
                fi_endian = fi.read(2).decode('utf-8')
            except Exception as e:
                raise SarpyIOError('Failed decoding the 2 character tiff header with error\n\t{}'.format(e))

            if fi_endian == 'II':
                self._endian = '<'
            elif fi_endian == 'MM':
                self._endian = '>'
            else:
                raise SarpyIOError('Invalid tiff endian string {}'.format(fi_endian))
            # check the magic number
            self._magic_number = numpy.fromfile(fi, dtype='{}i2'.format(self._endian), count=1)[0]
            if self._magic_number not in [42, 43]:
                raise SarpyIOError('Not a valid tiff file, got magic number {}'.format(self._magic_number))

            if self._magic_number == 43:
                rhead = numpy.fromfile(fi, dtype='{}i2'.format(self._endian), count=2)
                if rhead[0] != 8:
                    raise SarpyIOError('Not a valid bigtiff. The offset size is given as {}'.format(rhead[0]))
                if rhead[1] != 0:
                    raise SarpyIOError('Not a valid bigtiff. The reserved entry of '
                                  'the header is given as {} != 0'.format(rhead[1]))
        self._file_name = file_name
        self._tags = None

    @property
    def file_name(self) -> str:
        """
        str: READ ONLY. The file name.
        """

        return self._file_name

    @property
    def endian(self) -> str:
        """
        str: READ ONLY. The numpy dtype style ``('>' = big, '<' = little)`` endian string for the tiff file.
        """

        return self._endian

    @property
    def tags(self) -> Dict[str, Union[str, numpy.ndarray]]:
        """
        Dict: READ ONLY. The tiff tags dictionary,
        provided that :meth:`parse_tags` has been called. This dictionary is
        of the form `{<tag name> : str|numpy.ndarray}`, even for those tags
        containing only a single entry (i.e. `count=1`).
        """

        if self._tags is None:
            self.parse_tags()
        return self._tags

    def parse_tags(self) -> None:
        """
        Parse the tags from the file, if desired. This sets the `tags` attribute.

        Returns
        -------
        None
        """

        if self._magic_number == 42:
            type_dtype = numpy.dtype('{}u2'.format(self._endian))
            count_dtype = numpy.dtype('{}u2'.format(self._endian))
            offset_dtype = numpy.dtype('{}u4'.format(self._endian))
            offset_size = 4
        elif self._magic_number == 43:
            type_dtype = numpy.dtype('{}u2'.format(self._endian))
            count_dtype = numpy.dtype('{}i8'.format(self._endian))
            offset_dtype = numpy.dtype('{}i8'.format(self._endian))
            offset_size = 8
        else:
            raise ValueError('Unrecognized magic number {}'.format(self._magic_number))

        with open(self._file_name, 'rb') as fi:
            # skip the basic header
            fi.seek(offset_size, os.SEEK_SET)
            # extract the tags information
            tags = {}
            self._parse_ifd(fi, tags, type_dtype, count_dtype, offset_dtype, offset_size)
        self._tags = tags

    def _read_tag(self,
                  fi: BinaryIO,
                  tiff_type: int,
                  num_tag: int,
                  count: int) -> Dict:
        """
        Parse the specific tag information.

        Parameters
        ----------
        fi
            The file type object
        tiff_type : int
            The tag data type identifier value
        num_tag : int
            The numeric tag identifier
        count : int
            The number of such tags

        Returns
        -------
        dict
        """

        # find which tags we belong to
        if num_tag in _BASELINE_TAGS:
            ext = 'BaselineTag'
            name = _BASELINE_TAGS[num_tag]
        elif num_tag in _EXTENSION_TAGS:
            ext = 'ExtensionTag'
            name = _EXTENSION_TAGS[num_tag]
        elif num_tag in _GEOTIFF_TAGS:
            ext = 'GeoTiffTag'
            name = _GEOTIFF_TAGS[num_tag]
        else:
            ext, name = None, None
        # Now extract from file based on type number
        dtype = self._DTYPES.get(int(tiff_type), None)
        if dtype is None:
            logger.warning(
                'Failed to extract tiff data type {},\n\t'
                'for {} - {}'.format(tiff_type, ext, name))
            return {'Value': None, 'Name': name, 'Extension': ext}
        if tiff_type == 2:  # ascii field - read directly and decode?
            val = fi.read(count)  # this will be a string for python 2, and we decode for python 3
            if not isinstance(val, str):
                val = val.decode('utf-8')
            # eliminate the null characters
            val = re.sub('\x00', '', val)
        elif tiff_type in [5, 10]:  # unsigned or signed rational
            val = numpy.fromfile(fi, dtype='{}{}'.format(self._endian, dtype), count=numpy.int64(2*count)).reshape((-1, 2))
        else:
            val = numpy.fromfile(fi, dtype='{}{}'.format(self._endian, dtype), count=count)
            if count == 1:
                val = val[0]
        return {'Value': val, 'Name': name, 'Extension': ext}

    def _parse_ifd(self,
                   fi: BinaryIO,
                   tags: dict,
                   type_dtype: Union[str, numpy.dtype],
                   count_dtype: Union[str, numpy.dtype],
                   offset_dtype: Union[str, numpy.dtype],
                   offset_size: int) -> None:
        """
        Recursively parses the tag data and populates a provided dictionary
        Parameters
        ----------
        fi
            The file type object
        tags : dict
            The tag data dictionary being populated
        type_dtype : str|numpy.dtype
            The data type for the data element - note that endian-ness is included
        count_dtype : str|numpy.dtype
            The data type for the number of directories - note that endian-ness is included
        offset_dtype : str|numpy.dtype
            The data type for the offset - note that endian-ness is included
        offset_size : int
            The size of the offset
        Returns
        -------
        None
        """

        nifd = numpy.fromfile(fi, dtype=offset_dtype, count=1)[0]
        if nifd == 0:
            return  # termination criterion

        fi.seek(nifd)
        num_entries = numpy.fromfile(fi, dtype=count_dtype, count=1)[0]
        for entry in range(int(num_entries)):
            num_tag, tiff_type = numpy.fromfile(fi, dtype=type_dtype, count=2)
            count = numpy.fromfile(fi, dtype=offset_dtype, count=1)[0]
            total_size = self._SIZES[tiff_type-1]*count
            if total_size <= offset_size:
                save_ptr = fi.tell() + offset_size  # we should advance past the entire block
                value = self._read_tag(fi, tiff_type, num_tag, count)
                fi.seek(save_ptr)
            else:
                offset = numpy.fromfile(fi, dtype=offset_dtype, count=1)[0]
                save_ptr = fi.tell()  # save our current spot
                fi.seek(offset)  # get to the offset location
                value = self._read_tag(fi, tiff_type, num_tag, count)  # read the tag value
                fi.seek(save_ptr)  # return to our location
            tags[value['Name']] = value['Value']
        self._parse_ifd(fi, tags, type_dtype, count_dtype, offset_dtype, offset_size)  # recurse

    def check_compression(self):
        """
        Check the Compression tag, and verify uncompressed.

        Returns
        -------
        None
        """

        if self.tags['Compression'] != 1:
            raise ValueError(
                'The file {} indicates some kind of tiff compression, and the sarpy API requirements '
                'do not presently support reading of compressed tiff files. Consider using gdal to '
                'translate this tiff to an uncompressed file via the command\n\t'
                '"gdal_translate -co TILED=no <input_file> <output_file>"')

    def check_tiled(self):
        """
        Check if the tiff file is tiled.

        Returns
        -------
        None
        """

        if 'TileLength' in self.tags or 'TileWidth' in self.tags:
            raise ValueError(
                'The file {} indicates that this is a tiled file, and the sarpy API requirements '
                'do not presently support reading of tiled tiff files. Consider using gdal to '
                'translate this tiff to a flat file via the command\n\t'
                '"gdal_translate -co TILED=no <input_file> <output_file>"')


class NativeTiffDataSegment(NumpyMemmapSegment):
    """
    Direct reading of data from tiff file, failing if compression is present.

    This is a very complex SAR specific implementation, and not general.
    """

    __slots__ = ('_tiff_details', )
    _SAMPLE_FORMATS = {
        1: 'u', 2: 'i', 3: 'f', 5: 'i', 6: 'f'}  # 5 and 6 are complex int/float

    def __init__(self,
                 tiff_details: Union[str, TiffDetails],
                 reverse_axes: Union[None, int, Sequence[int]] = None,
                 transpose_axes: Union[None, Tuple[int, ...]] = None):
        """
        If format function and format_dtype are not provided, then SAR specific
        (not necessarily general) choices will be made.

        Parameters
        ----------
        tiff_details : TiffDetails
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]
        """

        if isinstance(tiff_details, str):
            tiff_details = TiffDetails(tiff_details)
        if not isinstance(tiff_details, TiffDetails):
            raise TypeError('NativeTiffChipper input argument must be a filename '
                            'or TiffDetails object.')

        tiff_details.check_compression()
        tiff_details.check_tiled()

        self._tiff_details = tiff_details
        if isinstance(tiff_details.tags['SampleFormat'], numpy.ndarray):
            samp_form = tiff_details.tags['SampleFormat'][0]
        else:
            samp_form = tiff_details.tags['SampleFormat']
        if samp_form not in self._SAMPLE_FORMATS:
            raise ValueError('Invalid sample format {}'.format(samp_form))
        if isinstance(tiff_details.tags['BitsPerSample'], numpy.ndarray):
            bits_per_sample = tiff_details.tags['BitsPerSample'][0]
        else:
            bits_per_sample = tiff_details.tags['BitsPerSample']

        raw_bands = int(tiff_details.tags['SamplesPerPixel'])

        if samp_form in [5, 6]:
            transform_data = 'COMPLEX'
            output_bands = int(raw_bands)
            raw_bands *= 2
            bits_per_sample /= 2
            output_dtype = 'complex64'
        elif raw_bands == 2:
            # NB: this is heavily skewed towards SAR and obviously not general
            transform_data = 'COMPLEX'
            output_dtype = 'complex64'
            output_bands = 1
        else:
            transform_data = None
            output_bands = raw_bands
            output_dtype = None

        raw_shape = (int(tiff_details.tags['ImageLength']), int(tiff_details.tags['ImageWidth']), raw_bands)
        raw_dtype = numpy.dtype('{0:s}{1:s}{2:d}'.format(
            self._tiff_details.endian, self._SAMPLE_FORMATS[samp_form], int(bits_per_sample/8)))
        if output_dtype is None:
            output_dtype = raw_dtype
        data_offset = int(tiff_details.tags['StripOffsets'][0])

        format_function = None
        if transform_data == 'COMPLEX':
            format_function = ComplexFormatFunction(raw_dtype, order='IQ')

        if reverse_axes is not None:
            if isinstance(reverse_axes, int):
                reverse_axes = (reverse_axes, )
            for entry in reverse_axes:
                if not entry < 2:
                    raise ValueError('reversing of axes on permitted along the first two axes.')

        if transpose_axes is not None:
            if len(transpose_axes) < 2 or len(transpose_axes) > 3:
                raise ValueError('transpose axes must have length 2 or 3')
            elif len(transpose_axes) == 2:
                transpose_axes = transpose_axes + (2, )

            if transpose_axes[2] != 2:
                raise ValueError(
                    'The transpose operation must preserve the location of the band data,\n\t'
                    'in the final dimension')

        if transpose_axes is None or transpose_axes == (0, 1, 2):
            output_shape = raw_shape[:2]
        else:
            output_shape = (raw_shape[1], raw_shape[0])

        if output_bands > 1:
            output_shape = output_shape + (output_bands, )

        NumpyMemmapSegment.__init__(
            self, tiff_details.file_name, data_offset, raw_dtype, raw_shape,
            formatted_dtype=output_dtype, formatted_shape=output_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode='r', close_file=True)

    @property
    def tiff_details(self) -> TiffDetails:
        return self._tiff_details


class TiffReader(BaseReader):
    def __init__(self,
                 tiff_details: Union[str, TiffDetails],
                 reverse_axes: Union[None, int, Sequence[int]] = None,
                 transpose_axes: Union[None, Tuple[int, ...]] = None):
        """

        Parameters
        ----------
        tiff_details : TiffDetails
        reverse_axes : None|int|Sequence[int]
        transpose_axes : None|Tuple[int, ...]
        """

        data_segment = NativeTiffDataSegment(tiff_details, reverse_axes=reverse_axes, transpose_axes=transpose_axes)
        BaseReader.__init__(self, data_segment, reader_type='OTHER', close_segments=True)

    @property
    def data_segment(self) -> NativeTiffDataSegment:
        """
        NativeTiffDataSegment: The tiff data segment.
        """

        return self._data_segment

    @property
    def tiff_details(self) -> TiffDetails:
        """
        TiffDetails: The tiff details object.
        """

        return self.data_segment.tiff_details

    @property
    def file_name(self):
        return self.tiff_details.file_name

########
# base expected functionality for a module with an implemented Reader


def is_a(file_name: str) -> Union[None, TiffReader]:
    """
    Tests whether a given file_name corresponds to a tiff file. Returns a
    tiff reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    None|TiffReader
        `TiffReader` instance if tiff file, `None` otherwise
    """

    try:
        tiff_details = TiffDetails(file_name)
        logger.info('File {} is determined to be a tiff file.'.format(file_name))
        return TiffReader(tiff_details)
    except SarpyIOError:
        # we don't want to catch parsing errors, for now
        return None




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class SENBAND1(TREElement):
    def __init__(self, value):
        super(SENBAND1, self).__init__()
        self.add_field('SENBAND', 's', 10, value)
        self.add_field('SEN_BANDWL', 's', 3, value)
        self.add_field('SEN_CEN_F', 's', 3, value)
        self.add_field('POLARIZATION', 's', 2, value)
        self.add_field('AZ_BWIDTH', 's', 6, value)
        self.add_field('EL_BWIDTH', 's', 6, value)
        self.add_field('DYN_RNGE', 's', 4, value)
        self.add_field('SENCALFAC', 's', 15, value)


class SENBAND2(TREElement):
    def __init__(self, value):
        super(SENBAND2, self).__init__()
        self.add_field('SENBAND', 's', 10, value)
        self.add_field('SEN_FOV_T', 's', 3, value)
        self.add_field('SEN_FOV_T_U', 's', 1, value)
        self.add_field('SEN_IFOV_T', 's', 3, value)
        self.add_field('SEN_IFOV_T_U', 's', 1, value)
        self.add_field('SEN_FOV_CT', 's', 5, value)
        self.add_field('SEN_IFOV_CT', 's', 3, value)
        self.add_field('SEN_IFOV_CT_U', 's', 1, value)
        self.add_field('SEN_FOR_T', 's', 3, value)
        self.add_field('SEN_FOR_CT', 's', 3, value)
        self.add_field('SEN_L_WAVE', 's', 4, value)
        self.add_field('SEN_U_WAVE', 's', 4, value)
        self.add_field('SUBBANDS', 's', 3, value)
        self.add_field('SENFLENGTH', 's', 4, value)
        self.add_field('SENFNUM', 's', 4, value)
        self.add_field('LINESAMPLES', 's', 4, value)
        self.add_field('DETECTTYPE', 's', 12, value)
        self.add_field('POLARIZATION', 's', 2, value)
        self.add_field('DYN_RNGE', 's', 4, value)
        self.add_field('SENCALFAC', 's', 15, value)


class SNSRAType(TREElement):
    def __init__(self, value):
        super(SNSRAType, self).__init__()
        self.add_field('VERNUM', 's', 4, value)
        self.add_field('SENNAME', 's', 20, value)
        self.add_field('SENTYPE', 's', 1, value)
        self.add_field('SENMODE', 's', 10, value)
        self.add_field('SENSCAN', 's', 12, value)
        self.add_field('SENSOR_ID', 's', 10, value)
        self.add_field('MPLAN', 's', 3, value)
        self.add_field('SENSERIAL', 's', 4, value)
        self.add_field('SENOPORG', 's', 10, value)
        self.add_field('SENMFG', 's', 12, value)
        self.add_field('ABSWVER', 's', 7, value)
        self.add_field('AVG_ALT', 's', 5, value)
        if self.SENTYPE == 'R':
            self.add_field('FOC_X', 's', 7, value)
            self.add_field('FOC_Y', 's', 7, value)
            self.add_field('FOC_Z', 's', 7, value)
            self.add_field('NUM_SENBAND', 'd', 1, value)
            self.add_loop('SENBANDs', self.NUM_SENBAND, SENBAND1, value)
        if self.SENTYPE in ['I', 'E']:
            self.add_field('NUM_SENBAND', 'd', 1, value)
            self.add_loop('SENBANDs', self.NUM_SENBAND, SENBAND2, value)


class SNSRA(TREExtension):
    _tag_value = 'SNSRA'
    _data_type = SNSRAType



"""
The ProductCreationType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import datetime
import logging
from typing import Union

import numpy

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import SerializableDescriptor, IntegerDescriptor, \
    StringDescriptor, StringEnumDescriptor, DateTimeDescriptor, ParametersDescriptor
from sarpy.io.complex.sicd_elements.SICD import SICDType

from .base import DEFAULT_STRICT


logger = logging.getLogger(__name__)


def extract_classification_from_sicd(sicd):
    """
    Extract a SIDD style classification from a SICD classification string.

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    str
    """

    if not isinstance(sicd, SICDType):
        raise TypeError('Requires SICDType instance, got type {}'.format(type(sicd)))

    c_str = sicd.CollectionInfo.Classification.upper().split('//')[0].strip()

    clas = None
    if c_str.startswith('UNCLASS') or c_str == 'U':
        clas = 'U'
    elif c_str.startswith('CONF') or c_str == 'C':
        clas = 'C'
    elif c_str.startswith('TOP ') or c_str == 'TS':
        clas = 'TS'
    elif c_str.startswith('SEC') or c_str == 'S':
        clas = 'S'
    elif c_str == 'FOUO' or c_str.startswith('REST') or c_str == 'R':
        clas = 'R'
    else:
        logger.critical(
            'Unclear how to extract classification code for classification string {}.\n\t'
            'It will default to unclassified, and should be set appropriately.'.format(c_str))
    return clas


class ProcessorInformationType(Serializable):
    """
    Details regarding the processor.
    """
    _fields = ('Application', 'ProcessingDateTime', 'Site', 'Profile')
    _required = ('Application', 'ProcessingDateTime', 'Site')
    # descriptors
    Application = StringDescriptor(
        'Application', _required, strict=DEFAULT_STRICT,
        docstring='Name and version of the application used to create the image.')  # type: str
    ProcessingDateTime = DateTimeDescriptor(
        'ProcessingDateTime', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='Date and time the image creation application processed the image (UTC).')  # type: numpy.datetime64
    Site = StringDescriptor(
        'Site', _required, strict=DEFAULT_STRICT,
        docstring='The creation site of this SICD product.')  # type: str
    Profile = StringDescriptor(
        'Profile', _required, strict=DEFAULT_STRICT,
        docstring='Identifies what profile was used to create this SICD product.')  # type: str

    def __init__(self, Application=None, ProcessingDateTime=None, Site=None, Profile=None, **kwargs):
        """

        Parameters
        ----------
        Application : str
        ProcessingDateTime : numpy.datetime64|datetime|date|str
        Site : str
        Profile : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Application = Application
        self.ProcessingDateTime = ProcessingDateTime
        self.Site = Site
        self.Profile = Profile
        super(ProcessorInformationType, self).__init__(**kwargs)


class ProductClassificationType(Serializable):
    """
    The overall classification of the product.
    """
    _fields = (
        'DESVersion', 'resourceElement', 'createDate', 'compliesWith', 'ISMCATCESVersion',
        'classification', 'ownerProducer', 'SCIcontrols', 'SARIdentifier',
        'disseminationControls', 'FGIsourceOpen', 'FGIsourceProtected', 'releasableTo',
        'nonICmarkings', 'classifiedBy', 'compilationReason', 'derivativelyClassifiedBy',
        'classificationReason', 'nonUSControls', 'derivedFrom', 'declassDate',
        'declassEvent', 'declassException', 'typeOfExemptedSource', 'dateOfExemptedSource',
        'SecurityExtensions')
    _required = (
        'DESVersion', 'createDate', 'classification', 'ownerProducer', 'compliesWith',
        'ISMCATCESVersion')
    _collections_tags = {'SecurityExtensions': {'array': False, 'child_tag': 'SecurityExtension'}}
    _set_as_attribute = (
        'DESVersion', 'resourceElement', 'createDate', 'compliesWith', 'ISMCATCESVersion',
        'classification', 'ownerProducer', 'SCIcontrols', 'SARIdentifier',
        'disseminationControls', 'FGIsourceOpen', 'FGIsourceProtected', 'releasableTo',
        'nonICmarkings', 'classifiedBy', 'compilationReason', 'derivativelyClassifiedBy',
        'classificationReason', 'nonUSControls', 'derivedFrom', 'declassDate',
        'declassEvent', 'declassException', 'typeOfExemptedSource', 'dateOfExemptedSource')
    _child_xml_ns_key = {the_field: 'ism' for the_field in _fields if the_field != 'SecurityExtensions'}
    # Descriptor
    DESVersion = IntegerDescriptor(
        'DESVersion', _required, strict=DEFAULT_STRICT, default_value=13,
        docstring='The version number of the DES. Should there be multiple specified in an instance document '
                  'the one at the root node is the one that will apply to the entire document.')  # type: int
    createDate = StringDescriptor(
        'createDate', _required, strict=DEFAULT_STRICT,
        docstring='This should be a date of format :code:`YYYY-MM-DD`, but this is not checked.')  # type: str
    compliesWith = StringDescriptor(
        'compliesWith', _required, strict=DEFAULT_STRICT, default_value="USGov",
        docstring='The ISM.XML rule sets a document complies with.')  # type: Union[None, str]
    ISMCATCESVersion = StringDescriptor(
        'ISMCATCESVersion', _required, strict=DEFAULT_STRICT, default_value='201903',
        docstring='')  # type: Union[None, str]
    classification = StringEnumDescriptor(
        'classification', ('U', 'C', 'R', 'S', 'TS'), _required, strict=DEFAULT_STRICT,
        docstring='')  # type: str
    ownerProducer = StringDescriptor(
        'ownerProducer', _required, strict=DEFAULT_STRICT,  # default_value='USA',
        docstring='')  # type: str
    SCIcontrols = StringDescriptor(
        'SCIcontrols', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    SARIdentifier = StringDescriptor(
        'SARIdentifier', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    disseminationControls = StringDescriptor(
        'disseminationControls', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    FGIsourceOpen = StringDescriptor(
        'FGIsourceOpen', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    FGIsourceProtected = StringDescriptor(
        'FGIsourceProtected', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    releasableTo = StringDescriptor(
        'releasableTo', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    nonICmarkings = StringDescriptor(
        'nonICmarkings', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    classifiedBy = StringDescriptor(
        'classifiedBy', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    compilationReason = StringDescriptor(
        'compilationReason', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    derivativelyClassifiedBy = StringDescriptor(
        'derivativelyClassifiedBy', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    classificationReason = StringDescriptor(
        'classificationReason', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    nonUSControls = StringDescriptor(
        'nonUSControls', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    derivedFrom = StringDescriptor(
        'derivedFrom', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    declassDate = StringDescriptor(
        'declassDate', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    declassEvent = StringDescriptor(
        'declassEvent', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    declassException = StringDescriptor(
        'declassException', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    typeOfExemptedSource = StringDescriptor(
        'typeOfExemptedSource', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    dateOfExemptedSource = StringDescriptor(
        'dateOfExemptedSource', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, str]
    SecurityExtensions = ParametersDescriptor(
        'SecurityExtensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Extensible parameters used to support profile-specific needs related to '
                  'product security.')  # type: ParametersCollection

    def __init__(self, DESVersion=13, createDate=None, compliesWith='USGov', ISMCATCESVersion='201903',
                 classification='U', ownerProducer='USA', SCIcontrols=None, SARIdentifier=None,
                 disseminationControls=None, FGIsourceOpen=None, FGIsourceProtected=None, releasableTo=None,
                 nonICmarkings=None, classifiedBy=None, compilationReason=None, derivativelyClassifiedBy=None,
                 classificationReason=None, nonUSControls=None, derivedFrom=None, declassDate=None,
                 declassEvent=None, declassException=None, typeOfExemptedSource=None, dateOfExemptedSource=None,
                 SecurityExtensions=None, **kwargs):
        """

        Parameters
        ----------
        DESVersion : int
        createDate : str
        compliesWith : None|str
        ISMCATCESVersion : None|str
        classification : str
        ownerProducer : str
        SCIcontrols : None|str
        SARIdentifier : None|str
        disseminationControls : None|str
        FGIsourceOpen : None|str
        FGIsourceProtected : None|str
        releasableTo : None|str
        nonICmarkings : None|str
        classifiedBy : None|str
        compilationReason : None|str
        derivativelyClassifiedBy : None|str
        classificationReason : None|str
        nonUSControls : None|str
        derivedFrom : None|str
        declassDate : None|str
        declassEvent : None|str
        declassException : None|str
        typeOfExemptedSource : None|str
        dateOfExemptedSource : None|str
        SecurityExtensions : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DESVersion = DESVersion
        self.createDate = createDate
        self.compliesWith = compliesWith
        self.ISMCATCESVersion = ISMCATCESVersion
        self.classification = classification
        self.ownerProducer = ownerProducer
        self.SCIcontrols = SCIcontrols
        self.SARIdentifier = SARIdentifier
        self.disseminationControls = disseminationControls
        self.FGIsourceOpen = FGIsourceOpen
        self.FGIsourceProtected = FGIsourceProtected
        self.releasableTo = releasableTo
        self.nonICmarkings = nonICmarkings
        self.classifiedBy = classifiedBy
        self.compilationReason = compilationReason
        self.derivativelyClassifiedBy = derivativelyClassifiedBy
        self.classificationReason = classificationReason
        self.nonUSControls = nonUSControls
        self.derivedFrom = derivedFrom
        self.declassDate = declassDate
        self.declassEvent = declassEvent
        self.declassException = declassException
        self.typeOfExemptedSource = typeOfExemptedSource
        self.dateOfExemptedSource = dateOfExemptedSource
        self.SecurityExtensions = SecurityExtensions
        super(ProductClassificationType, self).__init__(**kwargs)

    @property
    def resourceElement(self):
        return 'true'

    @classmethod
    def from_sicd(cls, sicd, create_date=None):
        """
        Extract best guess from SICD.

        Parameters
        ----------
        sicd : SICDType
        create_date : str

        Returns
        -------
        ProductClassificationType
        """

        clas = extract_classification_from_sicd(sicd)

        if create_date is None:
            create_date = datetime.datetime.now(tz=datetime.timezone.utc).date().isoformat()

        return cls(classification=clas, createDate=create_date)


class ProductCreationType(Serializable):
    """
    Contains general information about product creation.
    """

    _fields = (
        'ProcessorInformation', 'Classification', 'ProductName', 'ProductClass',
        'ProductType', 'ProductCreationExtensions')
    _required = (
        'ProcessorInformation', 'Classification', 'ProductName', 'ProductClass')
    _collections_tags = {'ProductCreationExtensions': {'array': False, 'child_tag': 'ProductCreationExtension'}}
    # Descriptors
    ProcessorInformation = SerializableDescriptor(
        'ProcessorInformation', ProcessorInformationType, _required, strict=DEFAULT_STRICT,
        docstring='Details regarding processor.')  # type: ProcessorInformationType
    Classification = SerializableDescriptor(
        'Classification', ProductClassificationType, _required, strict=DEFAULT_STRICT,
        docstring='The overall classification of the product.')  # type: ProductClassificationType
    ProductName = StringDescriptor(
        'ProductName', _required, strict=DEFAULT_STRICT,
        docstring='The output product name defined by the processor.')  # type: str
    ProductClass = StringDescriptor(
        'ProductClass', _required, strict=DEFAULT_STRICT,
        docstring='Class of product. Examples - :code:`Dynamic Image, Amplitude Change Detection, '
                  'Coherent Change Detection`')  # type: str
    ProductType = StringDescriptor(
        'ProductType', _required, strict=DEFAULT_STRICT,
        docstring='Type of sub-product. Examples - :code:`Frame #, Reference, Match`. '
                  'This field is only needed if there is a suite of associated '
                  'products.')  # type: Union[None, str]
    ProductCreationExtensions = ParametersDescriptor(
        'ProductCreationExtensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Extensible parameters used to support profile-specific needs related to '
                  'product creation.')  # type: ParametersCollection

    def __init__(self, ProcessorInformation=None, Classification=None, ProductName=None,
                 ProductClass=None, ProductType=None, ProductCreationExtensions=None, **kwargs):
        """

        Parameters
        ----------
        ProcessorInformation : ProcessorInformationType
        Classification : ProductClassificationType
        ProductName : str
        ProductClass : str
        ProductType : str
        ProductCreationExtensions : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ProcessorInformation = ProcessorInformation
        self.Classification = Classification
        self.ProductName = ProductName
        self.ProductClass = ProductClass
        self.ProductType = ProductType
        self.ProductCreationExtensions = ProductCreationExtensions
        super(ProductCreationType, self).__init__(**kwargs)

    @classmethod
    def from_sicd(cls, sicd, product_class):
        """
        Generate from a SICD for the given product class.

        Parameters
        ----------
        sicd : SICDType
        product_class : str

        Returns
        -------
        ProductCreationType
        """

        if not isinstance(sicd, SICDType):
            raise TypeError('Requires SICDType instance, got type {}'.format(type(sicd)))

        from sarpy.__about__ import __title__, __version__

        # use naive datetime because numpy warns about parsing timezone aware
        now = numpy.datetime64(datetime.datetime.now(tz=datetime.timezone.utc).replace(tzinfo=None))
        proc_info = ProcessorInformationType(
            Application='{} {}'.format(__title__, __version__),
            ProcessingDateTime=now,
            Site='Unknown')
        classification = ProductClassificationType.from_sicd(sicd)
        return cls(ProcessorInformation=proc_info,
                   Classification=classification,
                   ProductName=product_class,
                   ProductClass=product_class)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PIAEQAType(TREElement):
    def __init__(self, value):
        super(PIAEQAType, self).__init__()
        self.add_field('EQPCODE', 's', 7, value)
        self.add_field('EQPNOMEN', 's', 45, value)
        self.add_field('EQPMAN', 's', 64, value)
        self.add_field('OBTYPE', 's', 1, value)
        self.add_field('ORDBAT', 's', 3, value)
        self.add_field('CTRYPROD', 's', 2, value)
        self.add_field('CTRYDSN', 's', 2, value)
        self.add_field('OBJVIEW', 's', 6, value)


class PIAEQA(TREExtension):
    _tag_value = 'PIAEQA'
    _data_type = PIAEQAType



"""
Add a nominal noise polynomial to a sicd.

For a basic help on the command-line, check

>>> python -m sarpy.utils.nominal_sicd_noise --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import os
import argparse

import numpy

import sarpy
from sarpy.io.complex.sicd import SICDReader
from sarpy.io.complex.converter import conversion_utility
from sarpy.io.complex.sicd_elements.Radiometric import NoiseLevelType_

logger = logging.getLogger(__name__)


def nominal_sicd_noise(
        input_reader, out_directory, output_file=None, noise_db_value=-16.0, override=False,
        check_existence=True, check_older_version=False, preserve_nitf_information=False):
    """
    Create a sicd with the nominal noise value.

    Parameters
    ----------
    input_reader : str|SICDReader
    out_directory : str
        The directory of the output.
    output_file : None|str
        If `None`, then the name will mirror the original with noise details appended.
    noise_db_value : int|float
        The estimate for nesz in decibels.
    override : bool
        If a NoisePoly is already populated, should we override the value? If `False`
        and a NoisePoly is already populated, then an exception will be raised.
    check_existence : bool
        Check for the existence of the file before overwriting?
    check_older_version : bool
        Try to use a less recent version of SICD (1.1), for possible application
        compliance issues?
    preserve_nitf_information : bool
        Try to preserve some of the original NITF information?
    """

    if isinstance(input_reader, str):
        input_reader = SICDReader(input_reader)
    if not isinstance(input_reader, SICDReader):
        raise TypeError('We require that the input is a SICD reader or path to a sicd file.')

    noise_db_value = float(noise_db_value)
    if noise_db_value > -4:
        logger.warning(
            'The noise estimate should be provided in dB,\n\t'
            'and the provided value is `{}`.\n\t'
            'Maybe this is an error?'.format(noise_db_value))

    if output_file is None:
        fname = os.path.split(input_reader.file_name)[1]
        fstem, fext = os.path.splitext(fname)
        fstem += '_{}dB_noise'.format(int(noise_db_value))
        fname = fstem + fext
    else:
        fname = os.path.split(output_file)[1]

    sicd = input_reader.sicd_meta
    if sicd.Radiometric is None:
        raise ValueError(
            'The provided sicd does not contain any radiometric information,\n\t'
            'and SigmaZeroSFPoly is required to proceed')

    new_noise = NoiseLevelType_(
        NoisePoly=[[noise_db_value - 10 * numpy.log10(sicd.Radiometric.SigmaZeroSFPoly[0, 0]), ]],
        NoiseLevelType='ABSOLUTE')

    if sicd.Radiometric.NoiseLevel is None:
        original_noise = None
    else:
        if not override:
            raise ValueError(
                'The provided sicd already contains radiometric noise information,\n\t'
                'set override=True to replace the value')
        original_noise = sicd.Radiometric.NoiseLevel.copy()
    sicd.Radiometric.NoiseLevel = new_noise

    conversion_utility(
        input_reader, out_directory, output_files=fname,
        check_existence=check_existence,
        check_older_version=check_older_version,
        preserve_nitf_information=preserve_nitf_information)

    # return to the original noise information, so we haven't modified any in memory reader information
    sicd.Radiometric.NoiseLevel = original_noise


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Subset SICD file.",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        'input_file', metavar='input_file',
        help='Path input sicd data file.')
    parser.add_argument(
        'output_directory', metavar='output_directory',
        help='Path to the output directory. This directory MUST exist.')
    parser.add_argument(
        '--override', action='store_true',
        help='Override any present Noise polynomial?')
    parser.add_argument(
        '-o', '--output_file', default=None, type=str)
    parser.add_argument(
        '-n', '--noise', default=-16.0, type=float,
        help='Nominal nesz noise value in decibels')
    parser.add_argument(
        '-p', '--preserve', action='store_true',
        help='Try to preserve any NITF information?\n'
             'This only applies in the event that the file being read is a NITF')
    parser.add_argument(
        '-w', '--overwrite', action='store_true',
        help='Overwrite output file, if it already exists?')
    parser.add_argument(
        '--older', action='store_true',
        help='Try to use a less recent version of SICD (1.1),\n'
             'for possible application compliance issues?')
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose (level="INFO") logging?')

    args = parser.parse_args()

    level = 'INFO' if args.verbose else 'WARNING'
    logging.basicConfig(level=level)
    logger = logging.getLogger('sarpy')
    logger.setLevel(level)

    nominal_sicd_noise(
        args.input_file, args.output_directory, output_file=args.output_file,
        noise_db_value=args.noise,
        override=args.override,
        check_existence=not args.overwrite,
        check_older_version=args.older,
        preserve_nitf_information=args.preserve)



__classification__ = "UNCLASSIFIED"

import importlib.metadata


def entry_points(*, group):
    """
    Simple wrapper around importlib.metadata.entry_points

    Parameters
    ----------
    group : str
        entry point group name

    Returns
    -------
    list of entry points belonging to group

    Notes
    -----
    This function is only needed to support Python < 3.10.
    importlib.metadata was introduced in Python 3.8 as a provisional module.
    The stable interface was introduced in Python 3.10 and the original interface was removed in 3.12.
    """

    eps = importlib.metadata.entry_points()
    if hasattr(eps, 'select'):
        # Python >= 3.10
        return eps.select(group=group)
    else:
        return eps.get(group, [])


"""
Provides common methods for remapping a complex or other array to 8 or 16-bit
image type arrays.

Note: The original function and 8-bit implementation has been replaced with a
class based solution which allows state variables associated with the remap
function, and support for 16-bit versions, as well as an 8-bit MA, RGB or RGBA
lookup tables.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Wade Schwartzkopf", "Thomas McCullough")


import logging
from collections import OrderedDict
from typing import Dict, Union, Tuple, Type, List, Optional
import warnings

import numpy

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.utils import get_data_mean_magnitude, stats_calculation, \
    get_data_extrema

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None


logger = logging.getLogger(__name__)

_DEFAULTS_REGISTERED = False
_REMAP_DICT = OrderedDict()  # type: Dict[str, RemapFunction]

###########
# helper functions


def clip_cast(
        array: numpy.ndarray,
        dtype: Union[str, numpy.dtype, numpy.number] = 'uint8',
        min_value: Union[None, int, float] = None,
        max_value: Union[None, int, float] = None) -> numpy.ndarray:
    """
    Cast by clipping values outside of valid range, rather than truncating.

    Parameters
    ----------
    array : numpy.ndarray
    dtype : str|numpy.dtype
    min_value : None|int|float
    max_value : None|int|float

    Returns
    -------
    numpy.ndarray
    """

    np_type = numpy.dtype(dtype)
    min_value = numpy.iinfo(np_type).min if min_value is None else max(min_value, numpy.iinfo(np_type).min)
    max_value = numpy.iinfo(np_type).max if max_value is None else min(max_value, numpy.iinfo(np_type).max)
    return numpy.clip(array, min_value, max_value).astype(np_type)


def amplitude_to_density(
        data: numpy.ndarray,
        dmin: Union[int, float] = 30,
        mmult: Union[int, float] = 40,
        data_mean: Union[None, int, float] = None) -> numpy.ndarray:
    """
    Convert to density data for remap.

    This is a digested version of contents presented in a 1994 pulication
    entitled "Softcopy Display of SAR Data" by Kevin Mangis. It is unclear where
    this was first published or where it may be publicly available.

    Parameters
    ----------
    data : numpy.ndarray
        The (presumably complex) data to remap
    dmin : float|int
        A dynamic range parameter. Lower this widens the range, will raising it
        narrows the range. This was historically fixed at 30.
    mmult : float|int
        A contrast parameter. Low values will result is higher contrast and quicker
        saturation, while high values will decrease contrast and slower saturation.
        There is some balance between the competing effects in the `dmin` and `mmult`
        parameters.
    data_mean : None|float|int
        The data mean (for this or the parent array for continuity), which will
        be calculated if not provided.

    Returns
    -------
    numpy.ndarray
    """

    dmin = float(dmin)
    if not (0 <= dmin < 255):
        raise ValueError('Invalid dmin value {}'.format(dmin))

    mmult = float(mmult)
    if mmult < 1:
        raise ValueError('Invalid mmult value {}'.format(mmult))

    EPS = 1e-5
    amplitude = numpy.abs(data)
    if numpy.all(amplitude == 0):
        return amplitude
    else:
        if not data_mean:
            data_mean = numpy.mean(amplitude[numpy.isfinite(amplitude)])
        # remap parameters
        C_L = 0.8*data_mean
        C_H = mmult*C_L  # decreasing mmult will result in higher contrast (and quicker saturation)
        slope = (255 - dmin)/numpy.log10(C_H/C_L)
        constant = dmin - (slope*numpy.log10(C_L))
        # NB: C_H/C_L trivially collapses to mmult, but this is maintained for
        # clarity in historical reference
        # Originally, C_L and C_H were static values drawn from a determined set
        # of remap look-up tables. The C_L/C_H values were presumably based roughly
        # on mean amplitude and desired remap brightness/contrast. The dmin value
        # was fixed as 30.
        return slope*numpy.log10(numpy.maximum(amplitude, EPS)) + constant


def _linear_map(
        data: numpy.ndarray,
        min_value: float,
        max_value: float) -> numpy.ndarray:
    """
    Helper function which maps the input data, assumed to be of the correct from,
    into [0, 1] via a linear mapping (data - min_value)(max_value - min_value)
    and then clipping.

    Parameters
    ----------
    data : numpy.ndarray
    min_value : float
    max_value : float

    Returns
    -------
    numpy.ndarray
    """

    return numpy.clip((data - min_value)/float(max_value - min_value), 0, 1)


def _nrl_stats(
        amplitude: numpy.ndarray,
        percentile: Union[int, float] = 99) -> Tuple[float, float, float]:
    """
    Calculate the statistics for input into the nrl remap.

    Parameters
    ----------
    amplitude : numpy.ndarray
        The amplitude array, assumed real valued.
    percentile : float|int
        Which percentile to calculate

    Returns
    -------
    minimum: float
        Observed minimum
    maximum: float
        Observed maximum
    percentile: float
        Observed value with desired percentile
    """

    finite_mask = numpy.isfinite(amplitude)
    if numpy.any(finite_mask):
        return stats_calculation(amplitude[finite_mask], percentile=percentile)
    else:
        return 0., 0., 0.


############
# remap callable classes

class RemapFunction(object):
    """
    Abstract remap class which is callable.

    See the :func:`call` implementation for the given class to understand
    what specific keyword arguments are allowed for the specific instance.
    """

    _name = '_RemapFunction'
    __slots__ = ('_override_name', '_bit_depth', '_dimension')
    _allowed_dimension = {0, 1, 2, 3, 4}

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            dimension: int = 0):
        """

        Parameters
        ----------
        override_name : None|str
            Override name for a specific class instance
        bit_depth : int
            Should be one of 8 or 16
        dimension : int
        """

        self._override_name = None
        self._bit_depth = None
        self._dimension = None
        self._set_name(override_name)
        self._set_bit_depth(bit_depth)
        self._set_dimension(dimension)

    @property
    def name(self) -> str:
        """
        str: The (read-only) name for the remap function. This will be the
        override_name if one has been provided for this instance, otherwise it
        will be the generic `_name` class value.
        """

        return self._name if self._override_name is None else self._override_name

    def _set_name(self, value: Optional[str]):
        if value is None or isinstance(value, str):
            self._override_name = value
        else:
            raise ValueError('Got incompatible name')

    @property
    def bit_depth(self) -> int:
        """
        int: The (read-only) bit depth, which should be either 8 or 16.
        This is expected to be enforced by the implementation directly.
        """

        return self._bit_depth

    def _set_bit_depth(self, value: int):
        """
        This is intended to be read-only.

        Parameters
        ----------
        value : int
        """

        value = int(value)

        if value not in [8, 16, 32]:
            raise ValueError(
                'Bit depth is required to be one of 8, 16, or 32 and we got `{}`'.format(value))
        self._bit_depth = value

    @property
    def dimension(self) -> int:
        """
        int: The (read-only) size of the (additional) output final dimension.
        The value 0 is monochromatic, where the retuned output will have identical
        shape as input. Any other value should have additional final dimension of this size.
        """

        return self._dimension

    def _set_dimension(self, value: int):
        """
        The property is intended to be read-only.

        Parameters
        ----------
        value : int
        """

        value = int(value)
        if self._allowed_dimension is not None and value not in self._allowed_dimension:
            raise ValueError(
                'Dimension is required to be one of `{}`, got `{}`'.format(self._allowed_dimension, value))
        self._dimension = value

    @property
    def output_dtype(self) -> numpy.dtype:
        """
        numpy.dtype: The output data type.
        """

        if self._bit_depth == 8:
            return numpy.dtype('u1')
        elif self._bit_depth == 16:
            return numpy.dtype('u2')
        elif self._bit_depth == 32:
            return numpy.dtype('u4')
        else:
            raise ValueError('Unhandled bit_depth `{}`'.format(self._bit_depth))

    @property
    def are_global_parameters_set(self) -> bool:
        """
        bool: Are (all) global parameters used for applying this remap function
        set? This should return `True` if there are no global parameters.
        """

        return True

    def raw_call(
            self,
            data: numpy.ndarray,
            **kwargs) -> numpy.ndarray:
        """
        This performs the mapping from input data to output floating point
        version, this is directly used by the :func:`call` method.

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        kwargs
            Some keyword arguments may be allowed here

        Returns
        -------
        numpy.ndarray
            This should generally have `float64` dtype.
        """

        raise NotImplementedError

    def call(
            self,
            data: numpy.ndarray,
            **kwargs) -> numpy.ndarray:
        """
        This performs the mapping from input data to output discrete version.

        This method os directly called by the :func:`__call__` method, so the
        class instance (once constructed) is itself callable, as follows:

        >>> remap = RemapFunction()
        >>> discrete_data = remap(data, **kwargs)

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        kwargs
            Some keyword arguments may be allowed here

        Returns
        -------
        numpy.ndarray
        """

        return clip_cast(self.raw_call(data, **kwargs), dtype=self.output_dtype)

    def __call__(
            self,
            data: numpy.ndarray,
            **kwargs) -> numpy.ndarray:
        return self.call(data, **kwargs)

    @staticmethod
    def _validate_pixel_bounds(
            reader: SICDTypeReader,
            index: int,
            pixel_bounds: Union[None, Tuple, List, numpy.ndarray]):
        data_size = reader.get_data_size_as_tuple()[index]
        if pixel_bounds is None:
            return 0, data_size[0], 0, data_size[1]

        if not (
                (-data_size[0] <= pixel_bounds[0] <= data_size[0]) and
                (-data_size[0] <= pixel_bounds[1] <= data_size[0]) and
                (-data_size[1] <= pixel_bounds[2] <= data_size[1]) and
                (-data_size[1] <= pixel_bounds[3] <= data_size[1])):
            raise ValueError('invalid pixel bounds `{}` for data of shape `{}`'.format(pixel_bounds, data_size))
        return pixel_bounds

    def calculate_global_parameters_from_reader(
            self,
            reader: SICDTypeReader,
            index: int = 0,
            pixel_bounds: Union[None, Tuple, List, numpy.ndarray] = None):
        """
        Calculates any useful global bounds for the specified reader, the given
        index, and inside the given pixel bounds.

        This is expected to save ny necessary state here.

        Parameters
        ----------
        reader : SICDTypeReader
        index : int
        pixel_bounds : None|tuple|list|numpy.ndarray
            If provided, is of the form `(row min, row max, column min, column max)`.

        Returns
        -------
        None
        """

        raise NotImplementedError


class MonochromaticRemap(RemapFunction):
    """
    Abstract monochromatic remap class.
    """

    _name = '_Monochromatic'
    __slots__ = ('_override_name', '_bit_depth', '_dimension', '_max_output_value')
    _allowed_dimension = {0, }

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None):
        r"""

        Parameters
        ----------
        override_name : None|str
            Override name for a specific class instance
        bit_depth : int
        max_output_value : None|int
            The maximum output value. If provided, this must be in the interval
            :math:`[0, 2^{bit\_depth}]`
        """

        self._max_output_value = None
        RemapFunction.__init__(self, override_name=override_name, bit_depth=bit_depth, dimension=0)
        self._set_max_output_value(max_output_value)

    @property
    def max_output_value(self) -> int:
        """
        int: The (read-only) maximum output value size.
        """

        return self._max_output_value

    def _set_max_output_value(self, value: Optional[int]):
        max_possible = numpy.iinfo(self.output_dtype).max
        if value is None:
            value = max_possible
        else:
            value = int(value)

        if 0 < value <= max_possible:
            self._max_output_value = value
        else:
            raise ValueError(
                'the max_output_value must be between 0 and {}, '
                'got {}'.format(max_possible, value))

    def raw_call(self, data, **kwargs):
        raise NotImplementedError

    def calculate_global_parameters_from_reader(self, reader, index=0, pixel_bounds=None):
        raise NotImplementedError


############
# basic monchromatic collection

class Density(MonochromaticRemap):
    """
    A monochromatic logarithmic density remapping function.

    This is a digested version of contents presented in a 1994 publication
    entitled "Softcopy Display of SAR Data" by Kevin Mangis. It is unclear where
    this was first published or where it may be publicly available.
    """

    __slots__ = ('_override_name', '_bit_depth', '_dimension', '_dmin', '_mmult', '_eps', '_data_mean')
    _name = 'density'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            dmin: Union[float, int] = 30,
            mmult: Union[float, int] = 40,
            eps: float = 1e-5,
            data_mean: Union[None, int, float] = None):
        r"""

        Parameters
        ----------
        override_name : None|str
            Override name for a specific class instance
        bit_depth : int
        max_output_value : None|int
            The maximum output value. If provided, this must be in the interval
            :math:`[0, 2^{bit\_depth}]`
        dmin : float|int
            A dynamic range parameter. Lower this widens the range, will raising it
            narrows the range. This was historically fixed at 30.
        mmult : float|int
            A contrast parameter. Low values will result is higher contrast and quicker
            saturation, while high values will decrease contrast and slower saturation.
            There is some balance between the competing effects in the `dmin` and `mmult`
            parameters.
        eps : float
            small offset to create a nominal floor when mapping data containing 0's.
        data_mean : None|float|int
            The global data mean (for continuity). The appropriate value will be
            calculated on a per calling array basis if not provided.
        """

        MonochromaticRemap.__init__(self, override_name=override_name, bit_depth=bit_depth, max_output_value=max_output_value)
        self._data_mean = None
        self._dmin = None
        self._mmult = None
        self._eps = float(eps)

        self._set_dmin(dmin)
        self._set_mmult(mmult)
        self.data_mean = data_mean

    @property
    def dmin(self) -> float:
        """
        float: The dynamic range parameter. This is read-only.
        """

        return self._dmin

    def _set_dmin(self, value: Union[int, float]):
        value = float(value)
        if not (0 <= value < 255):
            raise ValueError('dmin must be in the interval [0, 255), got value {}'.format(value))
        self._dmin = value

    @property
    def mmult(self) -> float:
        """
        float: The contrast parameter. This is read only.
        """
        return self._mmult

    def _set_mmult(self, value: Union[int, float]):
        value = float(value)
        if value < 1:
            raise ValueError('mmult must be < 1, got {}'.format(value))
        self._mmult = value

    @property
    def data_mean(self) -> Optional[float]:
        """
        None|float: The data mean for global use.
        """
        return self._data_mean

    @data_mean.setter
    def data_mean(self, value: Optional[float]):
        if value is None:
            self._data_mean = None
            return

        self._data_mean = float(value)

    @property
    def are_global_parameters_set(self) -> bool:
        """
        bool: Is the global parameters used for applying this remap function
        set? In this case, this is the `data_mean` property.
        """

        return self._data_mean is not None

    def raw_call(
            self,
            data: numpy.ndarray,
            data_mean: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output floating point
        version, this is directly used by the :func:`call` method.

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        data_mean : None|float
            The pre-calculated data mean, for consistent global use. The order
            of preference is the value provided here, the class data_mean property
            value, then the value calculated from the present sample.

        Returns
        -------
        numpy.ndarray
        """

        data_mean = float(data_mean) if data_mean is not None else self._data_mean
        # the amplitude_to_density function is ort specifically geared towards
        # dynamic range of 0 - 255, just adjust it.
        multiplier = float(self.max_output_value)/255.0
        return multiplier*amplitude_to_density(data, dmin=self.dmin, mmult=self.mmult, data_mean=data_mean)

    def call(
            self,
            data: numpy.ndarray,
            data_mean: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output discrete version.

        This method os directly called by the :func:`__call__` method, so the
        class instance (once constructed) is itself callable, as follows:

        >>> remap = Density()
        >>> discrete_data = remap(data, data_mean=85.2)

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        data_mean : None|float
            The pre-calculated data mean, for consistent global use. The order
            of preference is the value provided here, the class data_mean property
            value, then the value calculated from the present sample.

        Returns
        -------
        numpy.ndarray
        """

        return clip_cast(
            self.raw_call(data, data_mean=data_mean),
            dtype=self.output_dtype, min_value=0, max_value=self.max_output_value)

    def calculate_global_parameters_from_reader(
            self,
            reader: SICDTypeReader,
            index: int = 0,
            pixel_bounds: Union[None, tuple, list, numpy.ndarray] = None):
        pixel_bounds = self._validate_pixel_bounds(reader, index, pixel_bounds)
        self.data_mean = get_data_mean_magnitude(pixel_bounds, reader, index, 25*1024*1024)


class Brighter(Density):
    """
    The density remap using parameters for brighter results.
    """

    _name = 'brighter'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            eps: float = 1e-5,
            data_mean: Union[None, int, float] = None):
        Density.__init__(
            self,
            override_name=override_name,
            bit_depth=bit_depth,
            max_output_value=max_output_value,
            dmin=60,
            mmult=40,
            eps=eps,
            data_mean=data_mean)


class Darker(Density):
    """
    The density remap using parameters for darker results.
    """

    _name = 'darker'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            eps: float = 1e-5,
            data_mean: Union[None, int, float] = None):
        Density.__init__(
            self,
            override_name=override_name,
            bit_depth=bit_depth,
            max_output_value=max_output_value,
            dmin=0,
            mmult=40,
            eps=eps,
            data_mean=data_mean)


class High_Contrast(Density):
    """
    The density remap using parameters for high contrast results.
    """

    _name = 'high_contrast'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            eps: float = 1e-5,
            data_mean: Union[None, int, float] = None):
        Density.__init__(
            self,
            override_name=override_name,
            bit_depth=bit_depth,
            max_output_value=max_output_value,
            dmin=30,
            mmult=4,
            eps=eps,
            data_mean=data_mean)


class GDM(MonochromaticRemap):
    """
    The Density remap using image specific parameters
    """

    __slots__ = ('_override_name', '_bit_depth', '_dimension', '_weighting', '_graze_rad', '_slope_rad',
                 '_data_mean', '_data_median')
    _name = 'gdm'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            weighting: Optional[str] = 'UNIFORM',
            graze_deg: Union[None, int, float] = None,
            slope_deg: Union[None, int, float] = None,
            data_mean: Union[None, int, float] = None,
            data_median: Union[None, int, float] = None
    ):
        """

        Parameters
        ----------
        override_name : None|str
            Override name for a specific class instance
        bit_depth : int
            Number of bits in unsigned integer output (default: 8)
        max_output_value: int
            Maximum output value [0, 2**bit_depth]
        weighting : str
            The spectral taper weighting ('taylor' | 'uniform') (default: 'uniform')
        graze_deg: None|float|int
            Required Graze angle (degrees)
        slope_deg: None|float|int
            Required Slope angle (degrees)
        data_mean : None|float|int
            The global data mean (for continuity). The appropriate value will be
            calculated on a per calling array basis if not provided.
        data_median : None|float|int
            The global data median (for continuity). The appropriate value will be
            calculated on a per calling array basis if not provided.
        """

        MonochromaticRemap.__init__(
            self, override_name=override_name, bit_depth=bit_depth, max_output_value=max_output_value)

        if graze_deg is None or slope_deg is None:
            raise ValueError("The Graze and Slope angles are required parameters of GDM remap.")

        self._data_mean = data_mean
        self._data_median = data_median
        self._weighting = weighting.lower()
        self._graze_rad = numpy.radians(graze_deg)
        self._slope_rad = numpy.radians(slope_deg)

    @property
    def data_mean(self) -> Optional[float]:
        """
        None|float: The data mean for global use.
        """
        return self._data_mean

    @data_mean.setter
    def data_mean(self, value: Optional[float]):
        if value is None:
            self._data_mean = None
            return

        self._data_mean = float(value)

    @property
    def data_median(self) -> Optional[float]:
        """
        None|float: The data median for global use.
        """
        return self._data_median

    @data_mean.setter
    def data_median(self, value: Optional[float]):
        self._data_median = None if value is None else float(value)

    def _cutoff_values(self, data_mean, data_median):
        # This is a subset of the GDM remap algorithm defined in the AGI Algorithm Description Document.
        # Only those parts of the algorithm relevant to existing SarPy capabilities are included here.
        # If a weighting name other than 'uniform' is specified, 'taylor' parameters will be chosen.
        c3 = {'taylor': 1.33, 'uniform': 1.23}.get(self._weighting, 1.33)
        w1 = {'taylor': 0.77, 'uniform': 0.77}.get(self._weighting, 0.77)
        w2 = -0.0422
        w5 = -1.95
        a1 = data_median
        a2 = numpy.sin(self._graze_rad) / numpy.cos(self._slope_rad)
        a5 = numpy.log10(data_median / data_mean)
        cl_init = a1 * w1
        ch_init = a1 * 10**(c3 + w2*a2 + w5*a5)
        r_min = 24
        r_max = 40
        r = 20*numpy.log10(ch_init / cl_init)
        if r < r_min:
            beta = 10**((r - r_min) / 40.0)         # pragma: nocover
        elif r > r_max:
            beta = 10**((r - r_max) / 40.0)         # pragma: nocover
        else:
            beta = 1.0
        cl = cl_init * beta
        ch = ch_init / beta

        return cl, ch

    def raw_call(
            self,
            data: numpy.ndarray,
            data_mean: Optional[float] = None,
            data_median: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output floating point
        version, this is directly used by the :func:`call` method.

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        data_mean : None|float
            The pre-calculated data mean, for consistent global use. The order
            of preference is the value provided here, the class data_mean property
            value, then the value calculated from the present sample.
        data_median : None|float
            The pre-calculated data median, for consistent global use. The order
            of preference is the value provided here, the class data_median property
            value, then the value calculated from the present sample.

        Returns
        -------
        numpy.ndarray
        """
        amplitude = numpy.abs(data)

        data_mean = float(data_mean) if data_mean is not None else self._data_mean
        if data_mean is None:
            # Warning: this is likely to be a local mean value rather than the global mean value.
            data_mean = numpy.mean(amplitude[numpy.isfinite(amplitude)])

        data_median = float(data_median) if data_median is not None else self._data_median
        if data_median is None:
            # Warning: this is likely to be a local median value rather than the global median value.
            data_median = numpy.median(amplitude[numpy.isfinite(amplitude)])

        c_l, c_h = self._cutoff_values(data_mean, data_median)

        multiplier = float(self.max_output_value)/255.0
        return multiplier*amplitude_to_density(data, dmin=30, mmult=c_h / c_l, data_mean=c_l / 0.8)

    def call(
            self,
            data: numpy.ndarray,
            data_mean: Optional[float] = None,
            data_median: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output discrete version.

        This method is directly called by the :func:`__call__` method, so the
        class instance (once constructed) is itself callable, as follows:

        >>> remap = GDM(weighting, graze_deg, slope_deg)
        >>> discrete_data = remap(data, data_mean, data_median)

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        data_mean : None|float
            The pre-calculated data mean, for consistent global use. The order
            of preference is the value provided here, the class data_mean property
            value, then the value calculated from the present sample.
        data_median : None|float
            The pre-calculated data median, for consistent global use. The order
            of preference is the value provided here, the class data_median property
            value, then the value calculated from the present sample.

        Returns
        -------
        numpy.ndarray
        """

        return clip_cast(
            self.raw_call(data, data_mean=data_mean, data_median=data_median),
            dtype=self.output_dtype, min_value=0, max_value=self.max_output_value)


class Linear(MonochromaticRemap):
    """
    A monochromatic linear remap function.
    """

    __slots__ = ('_override_name', '_bit_depth', '_dimension', '_max_value', '_min_value')
    _name = 'linear'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            min_value: Union[None, int, float] = None,
            max_value: Union[None, int, float] = None):
        """

        Parameters
        ----------
        override_name : None|str
            Override name for a specific class instance
        bit_depth : int
        min_value : None|float
        max_value : None|float
        """

        MonochromaticRemap.__init__(
            self,
            override_name=override_name,
            bit_depth=bit_depth,
            max_output_value=max_output_value)

        if min_value is not None:
            min_value = float(min_value)
        if max_value is not None:
            max_value = float(max_value)
        self._min_value = min_value
        self._max_value = max_value

    @property
    def min_value(self) -> Optional[float]:
        """
        None|float: The minimum value allowed (clipped below this)
        """
        return self._min_value

    @min_value.setter
    def min_value(self, value: Optional[float]):
        if value is None:
            self._min_value = None
        else:
            value = float(value)
            if not numpy.isfinite(value):
                raise ValueError('Got unsupported minimum value `{}`'.format(value))
            self._min_value = value

    @property
    def max_value(self) -> Optional[float]:
        """
        None|float:  The maximum value allowed (clipped above this)
        """

        return self._max_value

    @max_value.setter
    def max_value(self, value: Optional[float]):
        if value is None:
            self._max_value = None
        else:
            value = float(value)
            if not numpy.isfinite(value):
                raise ValueError('Got unsupported maximum value `{}`'.format(value))
            self._max_value = value

    @property
    def are_global_parameters_set(self) -> bool:
        """
        bool: Are (all) global parameters used for applying this remap function
        set? In this case, this is the `min_value` and `max_value` properties.
        """

        return self._min_value is not None and self._max_value is not None

    def _get_extrema(
            self,
            amplitude: numpy.ndarray,
            min_value: Optional[float],
            max_value: Optional[float]) -> Tuple[float, float]:
        if min_value is not None:
            min_value = float(min_value)
        if max_value is not None:
            max_value = float(max_value)

        if min_value is None:
            min_value = self.min_value
        if min_value is None:
            min_value = numpy.min(amplitude)

        if max_value is None:
            max_value = self.max_value
        if max_value is None:
            max_value = numpy.max(amplitude)

        # sanity check
        if min_value > max_value:
            min_value, max_value = max_value, min_value

        return min_value, max_value

    def raw_call(
            self,
            data: numpy.ndarray,
            min_value: Optional[float] = None,
            max_value: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output floating point
        version, this is directly used by the :func:`call` method.

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        min_value : None|float
            A minimum threshold, or pre-calculated data minimum, for consistent
            global use. The order of preference is the value provided here, the
            class `min_value` property value, then calculated from the present
            sample.
        max_value : None|float
            A maximum value threshold, or pre-calculated data maximum, for consistent
            global use. The order of preference is the value provided here, the
            class `max_value` property value, then calculated from the present
            sample.

        Returns
        -------
        numpy.ndarray
        """

        if numpy.iscomplexobj(data):
            amplitude = numpy.abs(data)
        else:
            amplitude = data

        out = numpy.empty(amplitude.shape, dtype='float64')
        max_output_value = self.max_output_value

        finite_mask = numpy.isfinite(amplitude)
        out[~finite_mask] = max_output_value

        if numpy.any(finite_mask):
            temp_data = amplitude[finite_mask]

            min_value, max_value = self._get_extrema(temp_data, min_value, max_value)

            if min_value == max_value:
                out[finite_mask] = 0
            else:
                out[finite_mask] = max_output_value*_linear_map(amplitude[finite_mask], min_value, max_value)
        return out

    def call(
            self,
            data: numpy.ndarray,
            min_value: Optional[float] = None,
            max_value: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output discrete version.

        This method os directly called by the :func:`__call__` method, so the
        class instance (once constructed) is itself callable, as follows:

        >>> remap = Linear()
        >>> discrete_data = remap(data, min_value=0, max_value=100)

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        min_value : None|float
            A minimum threshold, or pre-calculated data minimum, for consistent
            global use. The order of preference is the value provided here, the
            class `min_value` property value, then calculated from the present
            sample.
        max_value : None|float
            A maximum value threshold, or pre-calculated data maximum, for consistent
            global use. The order of preference is the value provided here, the
            class `max_value` property value, then calculated from the present
            sample.

        Returns
        -------
        numpy.ndarray
        """

        return clip_cast(
            self.raw_call(data, min_value=min_value, max_value=max_value),
            dtype=self.output_dtype, min_value=0, max_value=self.max_output_value)

    def calculate_global_parameters_from_reader(
            self,
            reader: SICDTypeReader,
            index: int = 0,
            pixel_bounds: Union[tuple, list, numpy.ndarray] = None) -> None:
        pixel_bounds = self._validate_pixel_bounds(reader, index, pixel_bounds)
        self.min_value, self.max_value = get_data_extrema(
            pixel_bounds, reader, index, 25*1024*1024, percentile=None)


class Logarithmic(MonochromaticRemap):
    """
    A logarithmic remap function.
    """

    __slots__ = ('_override_name', '_bit_depth', '_dimension', '_max_value', '_min_value')
    _name = 'log'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            min_value: Optional[float] = None,
            max_value: Optional[float] = None):
        """

        Parameters
        ----------
        override_name : None|str
            Override name for a specific class instance
        bit_depth : int
        min_value : None|float
        max_value : None|float
        """

        MonochromaticRemap.__init__(self, override_name=override_name, bit_depth=bit_depth, max_output_value=max_output_value)

        if min_value is not None:
            min_value = float(min_value)
        if max_value is not None:
            max_value = float(max_value)
        self._min_value = min_value
        self._max_value = max_value

    @property
    def min_value(self) -> Optional[float]:
        """
        None|float: The minimum value allowed (clipped below this)
        """
        return self._min_value

    @min_value.setter
    def min_value(self, value: Optional[float]):
        if value is None:
            self._min_value = None
        else:
            value = float(value)
            if not numpy.isfinite(value):
                raise ValueError('Got unsupported minimum value `{}`'.format(value))
            self._min_value = value

    @property
    def max_value(self) -> Optional[float]:
        """
        None|float:  The minimum value allowed (clipped above this)
        """

        return self._max_value

    @max_value.setter
    def max_value(self, value: Optional[float]):
        if value is None:
            self._max_value = None
        else:
            value = float(value)
            if not numpy.isfinite(value):
                raise ValueError('Got unsupported maximum value `{}`'.format(value))
            self._max_value = value

    @property
    def are_global_parameters_set(self) -> bool:
        """
        bool: Are (all) global parameters used for applying this remap function
        set? In this case, this is the `min_value` and `max_value` properties.
        """

        return self._min_value is not None and self._max_value is not None

    def _get_extrema(
            self,
            amplitude: numpy.ndarray,
            min_value: Optional[float],
            max_value: Optional[float]) -> Tuple[float, float]:
        if min_value is not None:
            min_value = float(min_value)
        if max_value is not None:
            max_value = float(max_value)

        if min_value is None:
            min_value = self.min_value
        if min_value is None:
            min_value = numpy.min(amplitude)

        if max_value is None:
            max_value = self.max_value
        if max_value is None:
            max_value = numpy.max(amplitude)

        # sanity check
        if min_value > max_value:
            min_value, max_value = max_value, min_value

        return min_value, max_value

    def raw_call(
            self,
            data: numpy.ndarray,
            min_value: Optional[float] = None,
            max_value: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output floating point
        version, this is directly used by the :func:`call` method.

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        min_value : None|float
            A minimum threshold, or pre-calculated data minimum, for consistent
            global use. The order of preference is the value provided here, the
            class `min_value` property value, then calculated from the present
            sample.
        max_value : None|float
            A maximum value threshold, or pre-calculated data maximum, for consistent
            global use. The order of preference is the value provided here, the
            class `max_value` property value, then calculated from the present
            sample.

        Returns
        -------
        numpy.ndarray
        """

        amplitude = numpy.abs(data)

        out = numpy.empty(amplitude.shape, dtype='float64')
        max_output_value = self.max_output_value

        finite_mask = numpy.isfinite(amplitude)
        zero_mask = (amplitude == 0)
        use_mask = finite_mask & (~zero_mask)

        out[~finite_mask] = max_output_value
        out[zero_mask] = 0

        if numpy.any(use_mask):
            temp_data = amplitude[use_mask]
            min_value, max_value = self._get_extrema(temp_data, min_value, max_value)

            if min_value == max_value:
                out[use_mask] = 0
            else:
                temp_data = (numpy.clip(temp_data, min_value, max_value) - min_value)/(max_value - min_value) + 1
                out[use_mask] = max_output_value*numpy.log2(temp_data)
        return out

    def call(
            self,
            data: numpy.ndarray,
            min_value: Optional[float] = None,
            max_value: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output discrete version.

        This method os directly called by the :func:`__call__` method, so the
        class instance (once constructed) is itself callable, as follows:

        >>> remap = Logarithmic()
        >>> discrete_data = remap(data, min_value=1.8, max_value=1.2e6)

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        min_value : None|float
            A minimum threshold, or pre-calculated data minimum, for consistent
            global use. The order of preference is the value provided here, the
            class `min_value` property value, then calculated from the present
            sample.
        max_value : None|float
            A maximum value threshold, or pre-calculated data maximum, for consistent
            global use. The order of preference is the value provided here, the
            class `max_value` property value, then calculated from the present
            sample.

        Returns
        -------
        numpy.ndarray
        """

        return clip_cast(
            self.raw_call(data, min_value=min_value, max_value=max_value),
            dtype=self.output_dtype, min_value=0, max_value=self.max_output_value)

    def calculate_global_parameters_from_reader(
            self,
            reader: SICDTypeReader,
            index: int = 0,
            pixel_bounds: Union[None, tuple, list, numpy.ndarray] = None) -> None:
        pixel_bounds = self._validate_pixel_bounds(reader, index, pixel_bounds)
        self.min_value, self.max_value = get_data_extrema(
            pixel_bounds, reader, index, 25*1024*1024, percentile=None)


class PEDF(MonochromaticRemap):
    """
    A monochromatic piecewise extended density format remap.
    """

    __slots__ = ('_override_name', '_bit_depth', '_dimension', '_density')
    _name = 'pedf'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            dmin: Union[int, float] = 30,
            mmult: Union[int, float] = 40,
            eps: float = 1e-5,
            data_mean: Union[None, int, float] = None):
        """

        Parameters
        ----------
        override_name : None|str
            Override name for a specific class instance
        bit_depth : int
        dmin : float|int
            A dynamic range parameter. Lower this widens the range, will raising it
            narrows the range. This was historically fixed at 30.
        mmult : float|int
            A contrast parameter. Low values will result is higher contrast and quicker
            saturation, while high values will decrease contrast and slower saturation.
            There is some balance between the competing effects in the `dmin` and `mmult`
            parameters.
        eps : float
            small offset to create a nominal floor when mapping data containing 0's.
        data_mean : None|float|int
            The global data mean (for continuity). The appropriate value will be
            calculated on a per calling array basis if not provided.
        """

        MonochromaticRemap.__init__(
            self, override_name=override_name, bit_depth=bit_depth, max_output_value=max_output_value)
        self._density = Density(
            bit_depth=bit_depth, max_output_value=max_output_value,
            dmin=dmin, mmult=mmult, eps=eps, data_mean=data_mean)

    @property
    def are_global_parameters_set(self) -> bool:
        """
        bool: Are (all) global parameters used for applying this remap function
        set? In this case, this is the `data_mean` property.
        """

        return self._density.are_global_parameters_set

    def raw_call(
            self,
            data: numpy.ndarray,
            data_mean: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output floating point
        version, this is directly used by the :func:`call` method.

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        data_mean : None|float
            The pre-calculated data mean, for consistent global use. The order
            of preference is the value provided here, the class data_mean property
            value, then the value calculated from the present sample.

        Returns
        -------
        numpy.ndarray
        """

        half_value = 0.5*self.max_output_value
        out = self._density.raw_call(data, data_mean=data_mean)
        top_mask = (out > half_value)
        out[top_mask] = 0.5*(out[top_mask] + half_value)
        return out

    def call(
            self,
            data: numpy.ndarray,
            data_mean: Optional[float] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output discrete version.

        This method os directly called by the :func:`__call__` method, so the
        class instance (once constructed) is itself callable, as follows:

        >>> remap = PEDF()
        >>> discrete_data = remap(data, data_mean=85.2)

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        data_mean : None|float
            The pre-calculated data mean, for consistent global use. The order
            of preference is the value provided here, the class data_mean property
            value, then the value calculated from the present sample.

        Returns
        -------
        numpy.ndarray
        """

        return clip_cast(
            self.raw_call(data, data_mean=data_mean),
            dtype=self.output_dtype, min_value=0, max_value=self.max_output_value)

    def calculate_global_parameters_from_reader(
            self,
            reader: SICDTypeReader,
            index: int = 0,
            pixel_bounds: Union[None, tuple, list, numpy.ndarray] = None) -> None:
        self._density.calculate_global_parameters_from_reader(
            reader, index=index, pixel_bounds=pixel_bounds)


class NRL(MonochromaticRemap):
    """
    A monochromatic remap which is linear for percentile of the data, then
    transitions to logarithmic.
    """

    __slots__ = ('_override_name', '_bit_depth', '_dimension', '_knee', '_percentile', '_stats')
    _name = 'nrl'

    def __init__(
            self,
            override_name: Optional[str] = None,
            bit_depth: int = 8,
            max_output_value: Optional[int] = None,
            knee: Optional[int] = None,
            percentile: Union[int, float] = 99,
            stats: Optional[Tuple[float, float, float]] = None):
        """
        Parameters
        ----------
        override_name : None|str
            Override name for a specific class instance
        bit_depth : int
        knee : int
            Where the knee for switching from linear to logarithmic occurs in the
            colormap regime - this should be in keeping with bit-depth.
        percentile : int|float
            In the event that we are calculating the stats, which percentile
            is the cut-off for lin-log switch-over?
        stats : None|tuple
            If provided, this should be of the form `(minimum, maximum, changeover)`.
        """

        self._knee = None
        self._percentile = None
        self._stats = None
        MonochromaticRemap.__init__(self, override_name=override_name, bit_depth=bit_depth, max_output_value=max_output_value)
        self._set_knee(knee)
        self._set_percentile(percentile)
        self._set_stats(stats)

    @property
    def knee(self) -> float:
        """
        float: The for switching from linear to logarithmic occurs in the colormap regime
        """

        return self._knee

    def _set_knee(self, knee: Optional[float]):
        max_value = self.max_output_value
        if knee is None:
            knee = 0.8*max_value
        knee = float(knee)
        if not (0 < knee < max_value):
            raise ValueError(
                'In keeping with bit-depth, knee must take a value strictly '
                'between 0 and {}'.format(max_value))
        self._knee = knee

    @property
    def percentile(self) -> float:
        """
        float: In the event that we are calculating the stats, which percentile
        is the cut-off for lin-log switch-over?
        """

        return self._percentile

    def _set_percentile(self, percentile: Union[None, int, float]):
        if percentile is None:
            percentile = 99.0
        else:
            percentile = float(percentile)

        if not (0 < percentile < 100):
            raise ValueError('percentile must fall strictly between 0 and 100')
        self._percentile = percentile

    @property
    def stats(self) -> Optional[Tuple[float, float, float]]:
        """
        None|tuple: If populated, this is a tuple of the form `(minimum, maximum, changeover)`.
        """

        return self._stats

    def _set_stats(self, value: Optional[Tuple[float, float, float]]):
        if value is None:
            self._stats = None
        else:
            self._stats = self._validate_stats(None, value)

    def _validate_stats(
            self,
            amplitude: Optional[numpy.ndarray],
            stats: Optional[Tuple[float, float, float]]) -> Optional[Tuple[float, float, float]]:
        if stats is None:
            stats = self.stats
        if stats is None and amplitude is not None:
            stats = _nrl_stats(amplitude, self.percentile)
        if stats is not None:
            min_value = float(stats[0])
            max_value = float(stats[1])
            changeover_value = float(stats[2])
            if not (min_value <= changeover_value <= max_value):
                raise ValueError('Got inconsistent stats value `{}`'.format(stats))
            stats = (min_value, max_value, changeover_value)
        return stats

    @property
    def are_global_parameters_set(self) -> bool:
        """
        bool: Are (all) global parameters used for applying this remap function
        set? In this case, this is the `stats` property.
        """

        return self._stats is not None

    def raw_call(
            self,
            data: numpy.ndarray,
            stats: Optional[Tuple[float, float, float]] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output floating point
        version, this is directly used by the :func:`call` method.

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        stats : None|tuple
            The stats `(minimum, maximum, chnageover)`, for consistent
            global use. The order of preference is the value provided here, the
            class `stats` property value, then calculated from the present
            sample.

        Returns
        -------
        numpy.ndarray
        """

        max_index = self.max_output_value

        amplitude = numpy.abs(data)
        amplitude_min, amplitude_max, changeover = self._validate_stats(amplitude, stats)
        out = numpy.empty(amplitude.shape, dtype='float64')
        if amplitude_min == amplitude_max:
            out[:] = 0
            return out

        linear_region = (amplitude <= changeover)
        if changeover > amplitude_min:

            out[linear_region] = numpy.clip(
                self.knee*_linear_map(amplitude[linear_region], amplitude_min, changeover),
                0,
                max_index)
        else:
            logger.warning(
                'The remap array is at least significantly constant, the nrl remap may return '
                'strange results.')
            out[linear_region] = 0

        if changeover == amplitude_max:
            out[~linear_region] = self.knee
        else:
            # calculate the log values
            extreme_data = numpy.clip(amplitude[~linear_region], changeover, amplitude_max)
            log_values = (extreme_data - changeover)/(amplitude_max - changeover) + 1
            # this is now linearly scaled from 1 to 2, apply log_2 and then scale appropriately
            out[~linear_region] = numpy.log2(log_values)*(max_index - self.knee) + self.knee
        return out

    def call(
            self,
            data: numpy.ndarray,
            stats: Optional[Tuple[float, float, float]] = None) -> numpy.ndarray:
        """
        This performs the mapping from input data to output discrete version.

        This method os directly called by the :func:`__call__` method, so the
        class instance (once constructed) is itself callable as follows:

        >>> remap = NRL()
        >>> discrete_data = remap(data, stats=(2.3, 1025.0, 997.2))

        Parameters
        ----------
        data : numpy.ndarray
            The (presumably) complex data to remap.
        stats : None|tuple
            The stats `(minimum, maximum, chnageover)`, for consistent
            global use. The order of preference is the value provided here, the
            class `stats` property value, then calculated from the present
            sample.

        Returns
        -------
        numpy.ndarray
        """

        return clip_cast(
            self.raw_call(data, stats=stats),
            dtype=self.output_dtype, min_value=0, max_value=self.max_output_value)

    def calculate_global_parameters_from_reader(
            self,
            reader: SICDTypeReader,
            index: int = 0,
            pixel_bounds: Union[None, tuple, list, numpy.ndarray] = None) -> None:
        pixel_bounds = self._validate_pixel_bounds(reader, index, pixel_bounds)
        self._set_stats(
            get_data_extrema(
                pixel_bounds, reader, index, 25*1024*1024, percentile=self.percentile))


class LUT8bit(RemapFunction):
    """
    A remap which uses a monochromatic remap function and an 8-bit lookup table
    to produce a (color) image output
    """

    __slots__ = ('_override_name', '_bit_depth', '_dimension', '_mono_remap', '_lookup_table')
    _name = '_lut_8bit'
    _allowed_dimension = None

    def __init__(
            self,
            mono_remap: MonochromaticRemap,
            lookup_table: Union[str, numpy.ndarray],
            override_name: Optional[str] = None,
            use_alpha: bool = False):
        """

        Parameters
        ----------
        mono_remap : MonochromaticRemap
            The remap to apply before using the lookup table. Note that the `max_output_value`
            and lookup_table first dimension size are required to be the same.
        lookup_table : str|numpy.ndarray
            A string name for a registered matplotlib colormap or the 256 element
            rgb or rgba array.
        override_name : None|str
            Override name for a specific class instance. If this is not provided and
            the `lookup_table` will be constructed from a matplotlib colormap name,
            then that name will be used.
        use_alpha : bool
            Only used if `mono_remap` is the name of a matplotlib colormap, this
            specifies whether or not to use the alpha channel.
        """

        self._mono_remap = None
        self._lookup_table = None
        if override_name is None and isinstance(lookup_table, str):
            override_name = lookup_table
        RemapFunction.__init__(
            self,
            override_name=override_name,
            bit_depth=8,
            dimension=0)
        # NB: dimension will be determined by the lookup table
        self._set_mono_remap(mono_remap)
        self._set_lookup_table(lookup_table, use_alpha)

    def _set_dimension(self, value: int):
        """
        The property is intended to be read-only.

        Parameters
        ----------
        value : int
        """

        self._dimension = value

    @property
    def mono_remap(self) -> MonochromaticRemap:
        """
        MonochromaticRemap: The monochromatic remap being used.
        """
        return self._mono_remap

    def _set_mono_remap(self, value: MonochromaticRemap):
        if not isinstance(value, MonochromaticRemap):
            raise ValueError('mono_remap requires a monochromatic remap instance')
        self._mono_remap = value

    @property
    def lookup_table(self) -> numpy.ndarray:
        """
        numpy.ndarray: The 8-bit lookup table.
        """

        return self._lookup_table

    def _set_lookup_table(
            self,
            value: Union[str, numpy.ndarray],
            use_alpha: bool) -> None:
        max_out_size = self.mono_remap.max_output_value
        if isinstance(value, str):
            if plt is None:
                raise ImportError(
                    'The lookup_table has been specified by providing a matplotlib '
                    'colormap name, but matplotlib can not be imported.')
            cmap = plt.get_cmap(value, max_out_size+1)
            color_array = cmap(numpy.arange(max_out_size+1))
            value = clip_cast(max_out_size*color_array, dtype='uint8')
            if value.shape[1] > 3 and not use_alpha:
                value = value[:, :3]
        if not (isinstance(value, numpy.ndarray) and value.ndim == 2 and value.dtype.name == 'uint8'):
            raise ValueError(
                'lookup_table requires a two-dimensional numpy array of dtype = uint8')
        if value.shape[0] != max_out_size+1:
            raise ValueError(
                'lookup_table size (first dimension) must agree with mono_remap.max_output_value')

        self._lookup_table = value
        self._dimension = value.shape[1]

    @property
    def are_global_parameters_set(self) -> bool:
        """
        bool: Are (all) global parameters used for applying this remap function set?
        """

        return self.mono_remap.are_global_parameters_set

    def raw_call(
            self,
            data: numpy.ndarray,
            **kwargs) -> numpy.ndarray:
        """
        Contrary to monochromatic remaps, this is identical to :func:`call`.

        Parameters
        ----------
        data : numpy.ndarray
        kwargs
            The keyword arguments passed through to mono_remap.

        Returns
        -------
        numpy.ndarray
        """

        return self._lookup_table[self._mono_remap(data, **kwargs)]

    def call(
            self,
            data: numpy.ndarray,
            **kwargs) -> numpy.ndarray:
        return self.raw_call(data, **kwargs)

    def calculate_global_parameters_from_reader(
            self,
            reader: SICDTypeReader,
            index: int = 0,
            pixel_bounds: Union[None, tuple, list, numpy.ndarray] = None) -> None:
        self.mono_remap.calculate_global_parameters_from_reader(
            reader, index=index, pixel_bounds=pixel_bounds)


###########
# registration function for maintaining the list

def register_remap(
        remap_function: Union[RemapFunction, Type],
        overwrite: bool = False) -> None:
    """
    Register a remap function for general usage.

    Parameters
    ----------
    remap_function : RemapFunction|Type
    overwrite : bool
        Should we overwrite any currently existing remap of the given name?

    Returns
    -------
    None
    """

    if isinstance(remap_function, type) and issubclass(remap_function, RemapFunction):
        remap_function = remap_function()
    if not isinstance(remap_function, RemapFunction):
        raise TypeError('remap_function must be an instance of RemapFunction.')

    remap_name = remap_function.name

    if remap_name not in _REMAP_DICT:
        _REMAP_DICT[remap_name] = remap_function
    elif overwrite:
        logger.info('Overwriting the remap {}'.format(remap_name))
        _REMAP_DICT[remap_name] = remap_function
    else:
        logger.info('Remap {} already exists and is not being replaced'.format(remap_name))


def _register_defaults():
    global _DEFAULTS_REGISTERED
    if _DEFAULTS_REGISTERED:
        return
    register_remap(NRL(bit_depth=8), overwrite=False)
    register_remap(Density(bit_depth=8), overwrite=False)
    register_remap(High_Contrast(bit_depth=8), overwrite=False)
    register_remap(Brighter(bit_depth=8), overwrite=False)
    register_remap(Darker(bit_depth=8), overwrite=False)
    register_remap(Linear(bit_depth=8), overwrite=False)
    register_remap(Logarithmic(bit_depth=8), overwrite=False)
    register_remap(PEDF(bit_depth=8), overwrite=False)
    if plt is not None:
        try:
            register_remap(LUT8bit(NRL(bit_depth=8), 'viridis', use_alpha=False), overwrite=False)
        except KeyError:
            pass
        try:
            register_remap(LUT8bit(NRL(bit_depth=8), 'magma', use_alpha=False), overwrite=False)
        except KeyError:
            pass
        try:
            register_remap(LUT8bit(NRL(bit_depth=8), 'rainbow', use_alpha=False), overwrite=False)
        except KeyError:
            pass
        try:
            register_remap(LUT8bit(NRL(bit_depth=8), 'bone', use_alpha=False), overwrite=False)
        except KeyError:
            pass

    _DEFAULTS_REGISTERED = True


def get_remap_names() -> List[str]:
    """
    Gets a list of currently registered remap function names.

    Returns
    -------
    List[str]
    """

    if not _DEFAULTS_REGISTERED:
        _register_defaults()
    return list(_REMAP_DICT.keys())


def get_remap_list() -> List[Tuple[str, RemapFunction]]:
    """
    Gets a list of currently registered remaps.

    Returns
    -------
    List[Tuple[str, RemapFunction]]
        List of tuples of the form `(<name>, <RemapFunction instance>)`.
    """

    if not _DEFAULTS_REGISTERED:
        _register_defaults()

    # NB: this was originally implemented via inspection of the callable members
    # of this module, but that ends up requiring more care in excluding
    # undesirable elements than this method
    return [(the_key, the_value) for the_key, the_value in _REMAP_DICT.items()]


def get_registered_remap(
        remap_name: str,
        default: Optional[RemapFunction] = None) -> RemapFunction:
    """
    Gets a remap function from it's registered name.

    Parameters
    ----------
    remap_name : str
    default : None|RemapFunction

    Returns
    -------
    RemapFunction

    Raises
    ------
    KeyError
    """

    if not _DEFAULTS_REGISTERED:
        _register_defaults()

    if remap_name in _REMAP_DICT:
        return _REMAP_DICT[remap_name]
    if default is not None:
        return default
    raise KeyError('Unregistered remap name `{}`'.format(remap_name))


#################
# DEPRECATED!
#################
# the original flat methods, maintained for a while
# for backwards compatibility

def density(data, data_mean=None):
    """
    Standard set of parameters for density remap.

    Parameters
    ----------
    data : numpy.ndarray
        The data to remap.
    data_mean : None|float|int

    Returns
    -------
    numpy.ndarray
    """

    warnings.warn(
        'the density() method is deprecated,\n\t'
        'use the Density class, which is also callable', DeprecationWarning)
    remapper = get_registered_remap('density')
    return remapper(data, data_mean=data_mean)


def brighter(data, data_mean=None):
    """
    Brighter set of parameters for density remap.

    Parameters
    ----------
    data : numpy.ndarray
    data_mean : None|float|int

    Returns
    -------
    numpy.ndarray
    """

    warnings.warn(
        'the brighter() method is deprecated,\n\t'
        'use the Brighter class, which is also callable', DeprecationWarning)
    remapper = get_registered_remap('brighter')
    return remapper(data, data_mean=data_mean)


def darker(data, data_mean=None):
    """
    Darker set of parameters for density remap.

    Parameters
    ----------
    data : numpy.ndarray
    data_mean : None|float|int

    Returns
    -------
    numpy.ndarray
    """

    warnings.warn(
        'the darker() method is deprecated,\n\t'
        'use the Darker class, which is also callable', DeprecationWarning)
    remapper = get_registered_remap('darker')
    return remapper(data, data_mean=data_mean)


def high_contrast(data, data_mean=None):
    """
    Increased contrast set of parameters for density remap.

    Parameters
    ----------
    data : numpy.ndarray
    data_mean : None|float|int

    Returns
    -------
    numpy.ndarray
    """

    warnings.warn(
        'the high_contrast() method is deprecated,\n\t'
        'use the HighContrast class, which is also callable', DeprecationWarning)
    remapper = get_registered_remap('high_contrast')
    return remapper(data, data_mean=data_mean)


def linear(data, min_value=None, max_value=None):
    """
    Linear remap - just the magnitude.

    Parameters
    ----------
    data : numpy.ndarray
    min_value : None|float
        The minimum allowed value for the dynamic range.
    max_value : None|float
        The maximum allowed value for the dynamic range.

    Returns
    -------
    numpy.ndarray
    """

    warnings.warn(
        'the linear() method is deprecated,\n\t'
        'use the Linear class, which is also callable', DeprecationWarning)
    remapper = get_registered_remap('linear')
    return remapper(data, min_value=min_value, max_value=max_value)


def log(data, min_value=None, max_value=None):
    """
    Logarithmic remap.

    Parameters
    ----------
    data : numpy.ndarray
    min_value : None|float
        The minimum allowed value for the dynamic range.
    max_value : None|float
        The maximum allowed value for the dynamic range.

    Returns
    -------
    numpy.ndarray
    """

    warnings.warn(
        'the log() method is deprecated,\n\t'
        'use the Logarithmic class, which is also callable', DeprecationWarning)
    remapper = get_registered_remap('log')
    return remapper(data, min_value=min_value, max_value=max_value)


def pedf(data, data_mean=None):
    """
    Piecewise extended density format remap.

    Parameters
    ----------
    data : numpy.ndarray
        The array to be remapped.
    data_mean : None|float|int

    Returns
    -------
    numpy.ndarray
    """

    warnings.warn(
        'the pedf() method is deprecated,\n\t'
        'use the PEDF class, which is also callable', DeprecationWarning)
    remapper = get_registered_remap('pedf')
    return remapper(data, data_mean=data_mean)


def nrl(data, stats=None):
    """
    A lin-log style remap.

    Parameters
    ----------
    data : numpy.ndarray
        The data array to remap
    stats : None|tuple
        This is calculated if not provided. Expected to be of the form
        `(minimum, maximum, 99th percentile)`.

    Returns
    -------
    numpy.ndarray
    """

    warnings.warn(
        'the nrl() method is deprecated,\n\t'
        'use the NRL class, which is also callable', DeprecationWarning)
    remapper = get_registered_remap('nrl')
    return remapper(data, stats=stats)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class OFFSETType(TREElement):
    def __init__(self, value):
        super(OFFSETType, self).__init__()
        self.add_field('LINE', 's', 8, value)
        self.add_field('SAMPLE', 's', 8, value)


class OFFSET(TREExtension):
    _tag_value = 'OFFSET'
    _data_type = OFFSETType



"""
The security tags definitions, which are used in each NITF subheader element.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from .base import NITFElement, _StringDescriptor, _StringEnumDescriptor, _parse_str


#########
# NITF 2.1 version

class NITFSecurityTags(NITFElement):
    """
    The NITF security tags object for NITF version 2.1 - see standards document
    Joint BIIF Profile (JBP) for more information.

    In the NITF standard, this object is simply redefined (is an identical way)
    for each of the main header and subheader objects. This object is intended to
    be flexibly used for any and all of these.
    """

    _ordering = (
        'CLAS', 'CLSY', 'CODE', 'CTLH', 'REL', 'DCTP', 'DCDT', 'DCXM',
        'DG', 'DGDT', 'CLTX', 'CAPT', 'CAUT', 'CRSN', 'SRDT', 'CTLN')
    _lengths = {
        'CLAS': 1, 'CLSY': 2, 'CODE': 11, 'CTLH': 2,
        'REL': 20, 'DCTP': 2, 'DCDT': 8, 'DCXM': 4,
        'DG': 1, 'DGDT': 8, 'CLTX': 43, 'CAPT': 1,
        'CAUT': 40, 'CRSN': 1, 'SRDT': 8, 'CTLN': 15}
    CLAS = _StringEnumDescriptor(
        'CLAS', True, 1, {'U', 'R', 'C', 'S', 'T'}, default_value='U',
        docstring='The classification level.')  # type: str
    CLSY = _StringDescriptor(
        'CLSY', True, 2, default_value='',
        docstring='Security Classification System. This field shall contain valid values '
                  'indicating the national or multinational security system used to classify this element. '
                  'Country Codes per FIPS PUB 10-4 shall be used to indicate national security systems. '
                  'The designator :code:`XN` is for classified data generated by a component using NATO security '
                  'system marking guidance. This code is outside the FIPS 10-4 document listing, and was '
                  'selected to not duplicate existing codes.')  # type: str
    CODE = _StringDescriptor(
        'CODE', True, 11, default_value='',
        docstring='Codewords. This field shall contain a valid indicator of the security '
                  'compartments associated. Valid values include one or more of the digraphs found '
                  'in table A-4. Multiple entries shall be separated by a single ECS spaces (0x20). '
                  'The selection of a relevant set of codewords is application '
                  'specific.')  # type: str
    CTLH = _StringDescriptor(
        'CTLH', True, 2, default_value='',
        docstring='Control and Handling. This field shall contain valid additional security control '
                  'and/or handling instructions (caveats) associated with this element.')  # type: str
    REL = _StringDescriptor(
        'REL', True, 20, default_value='',
        docstring='Releasing Instructions. This field shall contain a valid list of country and/or '
                  'multilateral entity codes to which countries and/or multilateral entities this element'
                  'is authorized for release. Valid items in the list are one or more country codes as '
                  'found in FIPS PUB 10-4 and/or codes identifying multilateral entities.')  # type: str
    DCTP = _StringEnumDescriptor(
        'DCTP', True, 2, {'', 'DD', 'DE', 'GD', 'GE', 'O', 'X'}, default_value='',
        docstring='Declassification Type. This field shall contain a valid indicator of the type of '
                  'security declassification or downgrading instructions which apply '
                  'to this element.')  # type: str
    DCDT = _StringDescriptor(
        'DCDT', True, 8, default_value='',
        docstring='Declassification Date. This field shall indicate the date on which this element '
                  'is to be declassified if the value in Declassification '
                  'Type is :code:`DD`.')  # type: str
    DCXM = _StringEnumDescriptor(
        'DCXM', True, 4,
        {'', 'X1', 'X2', 'X3', 'X4', 'X5', 'X6', 'X7', 'X8',
         '25X1', '25X2', '25X3', '25X4', '25X5', '25X6', '25X7', '25X8', '25X9',
         'DN10', 'DNI'}, default_value='',
        docstring='Declassification Exemption. This field shall indicate the reason this element is '
                  'exempt from automatic declassification if the value in Declassification '
                  'Type is :code:`X`.')  # type: str
    DG = _StringEnumDescriptor(
        'DG', True, 1, {'', 'S', 'C', 'R'}, default_value='',
        docstring='Downgrade. This field shall indicate the classification level to which this element is to '
                  'be downgraded if the values in Declassification Type are '
                  ':code:`GD` or :code:`GE`.')  # type: str
    DGDT = _StringDescriptor(
        'DGDT', True, 8, default_value='',
        docstring='Downgrade Date. This field shall indicate the date on which this element is to be downgraded '
                  'if the value in Declassification Type is :code:`GD`.')  # type: str
    CLTX = _StringDescriptor(
        'CLTX', True, 43, default_value='',
        docstring='Classification Text. This field shall be used to provide additional information about '
                  'classification to include identification of a declassification or downgrading event if the '
                  'values in Declassification Type are DE or GE. It may also be used to identify multiple '
                  'classification sources and/or any other special handling rules. '
                  'Values are user defined free text.')  # type: str
    CAPT = _StringEnumDescriptor(
        'CAPT', True, 1, {'', 'O', 'D', 'M'}, default_value='',
        docstring='Classification Authority Type. This field shall indicate the type of authority '
                  'used to classify this element.')  # type: str
    CAUT = _StringDescriptor(
        'CAUT', True, 40, default_value='',
        docstring='Classification Authority. This field shall identify the classification authority '
                  'for this element dependent upon the value in Classification Authority Type. Values are user '
                  'defined free text which should contain the following information: original classification '
                  'authority name and position or personal identifier if the value in Classification Authority '
                  'Type is O; title of the document or security classification guide used to classify this element '
                  'if the value in Classification Authority Type is D; and Derive-Multiple if the classification '
                  'was derived from multiple sources. In the latter case, the originator will maintain a record '
                  'of the sources used in accordance with existing security directives. One of the multiple '
                  'sources may also be identified in Classification Text if desired')  # type: str
    CRSN = _StringEnumDescriptor(
        'CRSN', True, 1, {'', 'A', 'B', 'C', 'D', 'E', 'F', 'G'}, default_value='',
        docstring='Classification Reason. This field shall contain values indicating the reason for '
                  'classifying the graphic. Valid values are A to G. These correspond to the reasons for '
                  'original classification per E.O. 12958, Section 1.5.(a) to (g).')  # type: str
    SRDT = _StringDescriptor(
        'SRDT', True, 8, default_value='',
        docstring='Security Source Date. This field shall indicate the date of the source used to derive '
                  'the classification of the graphic. In the case of multiple sources, the date of the '
                  'most recent source shall be used.')  # type: str
    CTLN = _StringDescriptor(
        'CTLN', True, 15, default_value='',
        docstring='Security Control Number. This field shall contain a valid security control number '
                  'associated with the graphic. The format of the security control number shall be in '
                  'accordance with the regulations governing the appropriate '
                  'security channel(s).')  # type: str


#########
# NITF 2.0 version

class NITFSecurityTags0(NITFElement):
    """
    The NITF security tags object for NITF version 2.0 - see standards document
    MIL-STD-2500A for more information.

    In the NITF standard, this object is simply redefined (is an identical way)
    for each of the main header and subheader objects. This object is intended to
    be flexibly used for any and all of these.
    """

    _ordering = (
        'CLAS', 'CODE', 'CTLH', 'REL', 'CAUT', 'CTLN', 'DWNG', 'DEVT')
    _lengths = {
        'CLAS': 1, 'CODE': 40, 'CTLH': 40, 'REL': 40, 'CAUT': 20,
        'CTLN': 20, 'DWNG': 6, 'DEVT': 40}
    CLAS = _StringEnumDescriptor(
        'CLAS', True, 1, {'U', 'R', 'C', 'S', 'T'}, default_value='U')  # type: str
    CODE = _StringDescriptor('CODE', True, 40)  # type: str
    CTLH = _StringDescriptor('CTLH', True, 40)  # type: str
    REL = _StringDescriptor('REL', True, 40)  # type: str
    CAUT = _StringDescriptor('CAUT', True, 20)  # type: str
    CTLN = _StringDescriptor('CTLN', True, 20)  # type: str
    DWNG = _StringDescriptor('DWNG', True, 6)  # type: str

    def __init__(self, **kwargs):
        self._DEVT = None
        super(NITFSecurityTags0, self).__init__(**kwargs)

    @property
    def DEVT(self):
        return self._DEVT

    @DEVT.setter
    def DEVT(self, value):
        self._DEVT = _parse_str(value, 40, None, 'DEVT', self)

    def _get_attribute_length(self, fld):
        if fld == 'DEVT':
            return 0 if self.DWNG != '999998' else self._lengths[fld]
        else:
            return super(NITFSecurityTags0, self)._get_attribute_length(fld)

    @classmethod
    def minimum_length(cls):
        # DEVT may not be there
        return super(NITFSecurityTags0, cls).minimum_length() - 40

    @classmethod
    def _parse_attribute(cls, fields, attribute, value, start):
        if attribute == 'DEVT':
            if fields['DWNG'] == b'999998':
                fields['DEVT'] = value[start:start+40].decode('utf-8')
                return start+40
            else:
                fields['DEVT'] = None
                # nothing to be done
                return start
        else:
            return super(NITFSecurityTags0, cls)._parse_attribute(fields, attribute, value, start)



"""
The DownstreamReprocessingType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import datetime
from typing import Union, List

import numpy

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import SerializableDescriptor, ParametersDescriptor, \
    StringDescriptor, DateTimeDescriptor, SerializableListDescriptor

from .base import DEFAULT_STRICT
from .blocks import RowColDoubleType, RowColIntType


class GeometricChipType(Serializable):
    """
    Contains information related to downstream chipping of the product. There is only one instance,
    and the instance is updated with respect to the **full image** parameters.

    For example, if an image is chipped out of a smaller chip, the new chip needs to be updated to
    the original full image corners. Since this relationship is linear, bi-linear interpolation is
    sufficient to determine an arbitrary chip coordinate in terms of the original full image coordinates.
    Chipping is typically done using an exploitation tool, and should not be done in the initial
    product creation.
    """

    _fields = (
        'ChipSize', 'OriginalUpperLeftCoordinate', 'OriginalUpperRightCoordinate',
        'OriginalLowerLeftCoordinate', 'OriginalLowerRightCoordinate')
    _required = _fields
    # Descriptor
    ChipSize = SerializableDescriptor(
        'ChipSize', RowColIntType, _required, strict=DEFAULT_STRICT,
        docstring='Size of the chipped product in pixels.')  # type: RowColIntType
    OriginalUpperLeftCoordinate = SerializableDescriptor(
        'OriginalUpperLeftCoordinate', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='Upper-left corner with respect to the original product.')  # type: RowColDoubleType
    OriginalUpperRightCoordinate = SerializableDescriptor(
        'OriginalUpperRightCoordinate', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='Upper-right corner with respect to the original product.')  # type: RowColDoubleType
    OriginalLowerLeftCoordinate = SerializableDescriptor(
        'OriginalLowerLeftCoordinate', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='Lower-left corner with respect to the original product.')  # type: RowColDoubleType
    OriginalLowerRightCoordinate = SerializableDescriptor(
        'OriginalLowerRightCoordinate', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='Lower-right corner with respect to the original product.')  # type: RowColDoubleType

    def __init__(self, ChipSize=None, OriginalUpperLeftCoordinate=None, OriginalUpperRightCoordinate=None,
                 OriginalLowerLeftCoordinate=None, OriginalLowerRightCoordinate=None, **kwargs):
        """

        Parameters
        ----------
        ChipSize : RowColIntType|numpy.ndarray|list|tuple
        OriginalUpperLeftCoordinate : RowColDoubleType|numpy.ndarray|list|tuple
        OriginalUpperRightCoordinate : RowColDoubleType|numpy.ndarray|list|tuple
        OriginalLowerLeftCoordinate : RowColDoubleType|numpy.ndarray|list|tuple
        OriginalLowerRightCoordinate : RowColDoubleType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ChipSize = ChipSize
        self.OriginalUpperLeftCoordinate = OriginalUpperLeftCoordinate
        self.OriginalUpperRightCoordinate = OriginalUpperRightCoordinate
        self.OriginalLowerLeftCoordinate = OriginalLowerLeftCoordinate
        self.OriginalLowerRightCoordinate = OriginalLowerRightCoordinate
        super(GeometricChipType, self).__init__(**kwargs)


class ProcessingEventType(Serializable):
    """
    Processing event data.
    """

    _fields = ('ApplicationName', 'AppliedDateTime', 'InterpolationMethod', 'Descriptors')
    _required = ('ApplicationName', 'AppliedDateTime')
    _collections_tags = {'Descriptors': {'array': False, 'child_tag': 'Descriptor'}}
    # Descriptor
    ApplicationName = StringDescriptor(
        'ApplicationName', _required, strict=DEFAULT_STRICT,
        docstring='Application which applied a modification.')  # type: str
    AppliedDateTime = DateTimeDescriptor(
        'AppliedDateTime', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='Date and time defined in Coordinated Universal Time (UTC).')  # type: numpy.datetime64
    InterpolationMethod = StringDescriptor(
        'InterpolationMethod', _required, strict=DEFAULT_STRICT,
        docstring='Type of interpolation applied to the data.')  # type: Union[None, str]
    Descriptors = ParametersDescriptor(
        'Descriptors', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Descriptors for the processing event.')  # type: ParametersCollection

    def __init__(self, ApplicationName=None, AppliedDateTime=None, InterpolationMethod=None,
                 Descriptors=None, **kwargs):
        """

        Parameters
        ----------
        ApplicationName : str
        AppliedDateTime : numpy.datetime64|str
        InterpolationMethod : None|str
        Descriptors : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        # use naive datetime because numpy warns about parsing timezone aware
        now = numpy.datetime64(datetime.datetime.now(tz=datetime.timezone.utc).replace(tzinfo=None))
        self.ApplicationName = ApplicationName
        self.AppliedDateTime = now if AppliedDateTime is None else AppliedDateTime
        self.InterpolationMethod = InterpolationMethod
        self.Descriptors = Descriptors
        super(ProcessingEventType, self).__init__(**kwargs)


class DownstreamReprocessingType(Serializable):
    """
    Further processing after initial image creation.
    """

    _fields = ('GeometricChip', 'ProcessingEvents')
    _required = ()
    _collections_tags = {'ProcessingEvents': {'array': False, 'child_tag': 'ProcessingEvent'}}
    # Descriptor
    GeometricChip = SerializableDescriptor(
        'GeometricChip', GeometricChipType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, GeometricChipType]
    ProcessingEvents = SerializableListDescriptor(
        'ProcessingEvents', ProcessingEventType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: Union[None, List[ProcessingEventType]]

    def __init__(self, GeometricChip=None, ProcessingEvents=None, **kwargs):
        """

        Parameters
        ----------
        GeometricChip : None|GeometricChipType
        ProcessingEvents : None|List[ProcessingEventType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.GeometricChip = GeometricChip
        self.ProcessingEvents = ProcessingEvents
        super(DownstreamReprocessingType, self).__init__(**kwargs)



"""
The ProductInfo elements.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List

import numpy

from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import StringDescriptor, DateTimeDescriptor, \
    ParametersDescriptor, SerializableListDescriptor
from .base import DEFAULT_STRICT


class CreationInfoType(Serializable):
    """
    Parameters that provide general information about the CPHD product generation.
    """

    _fields = ('Application', 'DateTime', 'Site', 'Parameters')
    _required = ('DateTime', )
    _collections_tags = {'Parameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    Application = StringDescriptor(
        'Application', _required, strict=DEFAULT_STRICT,
        docstring='Name and version of the application used to create the CPHD.')  # type: str
    DateTime = DateTimeDescriptor(
        'DateTime', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='Date and time the image creation application processed the image (UTC).')  # type: numpy.datetime64
    Site = StringDescriptor(
        'Site', _required, strict=DEFAULT_STRICT,
        docstring='The creation site of this CPHD product.')  # type: str
    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Additional parameters.')  # type: Union[None, ParametersCollection]

    def __init__(self, Application=None, DateTime=None, Site=None, Parameters=None, **kwargs):
        """

        Parameters
        ----------
        Application : str
        DateTime : numpy.datetime64|datetime|date|str
        Site : str
        Profile : str
        Parameters : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Application = Application
        self.DateTime = DateTime
        self.Site = Site
        self.Parameters = Parameters
        super(CreationInfoType, self).__init__(**kwargs)


class ProductInfoType(Serializable):
    """
    Parameters that provide general information about the CPHD product and/or the
    derived products that may be created from it.
    """

    _fields = ('Profile', 'CreationInfos', 'Parameters')
    _required = ()
    _collections_tags = {
        'CreationInfos': {'array': False, 'child_tag': 'CreationInfo'},
        'Parameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    Profile = StringDescriptor(
        'Profile', _required, strict=DEFAULT_STRICT,
        docstring='Identifies what profile was used to create this CPHD product.')  # type: str
    CreationInfos = SerializableListDescriptor(
        'CreationInfos', CreationInfoType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that provide general information about the CPHD '
                  'product generation.')  # type: Union[None, List[CreationInfoType]]
    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Additional parameters.')  # type: Union[None, ParametersCollection]

    def __init__(self, Profile=None, CreationInfos=None, Parameters=None, **kwargs):
        """

        Parameters
        ----------
        Profile : str
        CreationInfos : None|List[CreationInfoType]
        Parameters : None|ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Profile = Profile
        self.CreationInfos = CreationInfos
        self.Parameters = Parameters
        super(ProductInfoType, self).__init__(**kwargs)



"""
Convert from complex SAR image format to SICD format.

For a basic help on the command-line, check

>>> python -m sarpy.utils.convert_to_sicd --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Valkyrie Systems Corporation")

import argparse
import logging

import sarpy
from sarpy.io.complex.converter import conversion_utility


def convert(input_file, output_dir, preserve_nitf_information=False,
            dem_filename_pattern=None, dem_type=None, geoid_file=None):
    """

    Parameters
    ----------
    input_file : str
        Path to the input file.
    output_dir : str
        Output directory path.
    preserve_nitf_information : bool
        Try to preserve NITF information? This only applies in the case that the
        file being read is actually a NITF file.
    dem_filename_pattern : str | None
        Optional string specifying a Digital Elevation Model (DEM) filename pattern.
        This is a format string that specifies a glob pattern that will
        uniquely specify a DEM file from the Lat/Lon of the SW corner of
        the DEM tile.  See the convert_to_sicd help text for more details.
    dem_type : str | None
        Optional DEM type ('GeoTIFF', 'GeoTIFF:WGS84', 'GeoTIFF:EGM2008', etc.).
        This parameter is required when dem_filename_pattern is specified.  For 'GeoTIFF'
        DEM files, the reference surface can be either WGS84 or any of the geoid models.
        The reference surface is appended to the DEM type with a ':' separator.  If the
        reference surface is not specified, then EGM2008 is assumed.
    geoid_file : str | None
        Optional Geoid file which might be needed when dem_filename_pattern is specified.
    """

    conversion_utility(input_file, output_dir, preserve_nitf_information=preserve_nitf_information,
                       dem_filename_pattern=dem_filename_pattern, dem_type=dem_type, geoid_file=geoid_file)


if __name__ == '__main__':
    epilog = ('Note:\n'
              'The DEM files must have the SW corner Lat/Lon encoded in their filenames.\n'
              'The --dem-path-pattern argument contains a format string that when populated will\n'
              'create as glob pattern that will specify the desired DEM file.  The following\n'
              'arguments are provided to the format string.\n'
              '    lat = int(numpy.floor(lat))\n'
              '    lon = int(numpy.floor(lon))\n'
              '    abslat = int(abs(numpy.floor(lat)))\n'
              '    abslon = int(abs(numpy.floor(lon)))\n'
              '    ns = "s" if lat < 0 else "n"\n'
              '    NS = "S" if lat < 0 else "N"\n'
              '    ew = "w" if lon < 0 else "e"\n'
              '    EW = "W" if lon < 0 else "E"\n'
              '\n'
              'For example (with Linux file separators) the following specifies a GeoTIFF DEM pattern:\n'
              '    /dem_root/tdt_{ns}{abslat:02}{ew}{abslon:03}_*/DEM/TDT_{NS}{abslat:02}{EW}{abslon:03}_*_DEM.tif\n'
              '\n'
              'In theory, one could use a simple format string using wildcard characters like this:\n'
              '    /dem_root/**/*{NS}{abslat:02}{EW}{abslon:03}*DEM.tif\n'
              '\n'
              'However, this would be unwise since glob might have to scan through many files and directories\n'
              'to find the desired file.  This could be quite time consuming if there are many files in dem_root.\n'
              )

    parser = argparse.ArgumentParser(description="Convert to SICD format.",
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     epilog=epilog)
    parser.add_argument(
        'input_file', metavar='input_file',
        help='Path input data file, or directory for radarsat, RCM, sentinel or other systems.\n'
             '* For radarsat or RCM, this can be the product.xml file, or parent directory\n'
             '  of product.xml or metadata/product.xml.\n'
             '* For sentinel, this can be the manifest.safe file, or parent directory of\n'
             '  manifest.safe.\n')
    parser.add_argument(
        'output_directory', metavar='output_directory',
        help='Path to the output directory. This directory MUST exist.\n'
             '* Depending on the input details, multiple SICD files may be produced.\n'
             '* The name for the output file(s) will be chosen based on CoreName and\n '
             '  transmit/collect polarization.\n')
    parser.add_argument(
        '-p', '--preserve', action='store_true',
        help='Try to preserve any NITF information?\n'
             'This only applies in the event that the file being read is a NITF')
    parser.add_argument(
        '-d', '--dem-filename-pattern',
        help='Optional string specifying a Digital Elevation Model (DEM) filename pattern.\n'
             'This is a format string that specifies a glob pattern that will\n'
             'uniquely specify a DEM file from the Lat/Lon of the SW corner of\n'
             'the DEM tile.  See the note below for more details.\n')
    parser.add_argument(
        '-t', '--dem-type',
        help=('Optional DEM type ("GeoTIFF", etc.).\n'
              'This parameter is required when dem-path-pattern is specified.\n'))
    parser.add_argument(
        '-g', '--geoid-file',
        help='Optional path to a geoid definition file.\n'
             'A geoid definition file is required when dem-path-pattern is specified\n'
             'and the DEM height values are relative to a geoid.\n')
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose (level="INFO") logging?')

    args = parser.parse_args()
    level = 'INFO' if args.verbose else 'WARNING'
    logging.basicConfig(level=level)
    logger = logging.getLogger('sarpy')
    logger.setLevel(level)

    convert(args.input_file, args.output_directory, preserve_nitf_information=args.preserve,
            dem_filename_pattern=args.dem_filename_pattern, dem_type=args.dem_type, geoid_file=args.geoid_file)
